# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from datetime import datetime
from typing import Dict, Optional

import numpy as np
import requests

from geti_sdk.data_models import Prediction
from geti_sdk.deployment.inference_hook_interfaces import PostInferenceAction
from geti_sdk.rest_converters import PredictionRESTConverter


class HttpRequestAction(PostInferenceAction):
    """
    Post inference action that will send an http request to a specified `url`.

    :param url: Full URL to send the request to
    :param method: HTTP method to use, for example `GET` or `POST`
    :param headers: Headers to use in the request, can be used for instance for
        authentication purposes
    :param include_prediction_data: Set this to True to include the prediction data in
        the body of the request. Only applicable for POST requests
    :param log_level: Log level for the action. Options are 'info' or 'debug'
    """

    def __init__(
        self,
        url: str,
        method: str = "GET",
        headers: Optional[Dict[str, str]] = None,
        include_prediction_data: bool = False,
        log_level: str = "info",
    ):
        super().__init__(log_level=log_level)

        self.url = url
        self.method = method

        if headers is None:
            headers = {}
        self.headers = headers

        self.include_data = include_prediction_data and method == "POST"
        if self.include_data:
            self.headers.update({"Content-Type": "application/json"})

        self._repr_info_ = (
            f"url=`{url}`, "
            f"method={method}, "
            f"headers={headers}, "
            f"include_data={self.include_data}"
        )

    def __call__(
        self,
        image: np.ndarray,
        prediction: Prediction,
        score: Optional[float] = None,
        name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
    ):
        """
        Execute the action, send an HTTP request to the predefined url

        :param image: Numpy array representing an image
        :param prediction: Prediction object which was generated for the image
        :param score: Optional score computed from a post inference trigger
        :param name: String containing the name of the image
        :param timestamp: Datetime object containing the timestamp belonging to the
            image
        """
        data = None
        if self.include_data:
            prediction_dict = PredictionRESTConverter.to_dict(prediction)
            data = prediction_dict

        requests.request(
            method=self.method, url=self.url, headers=self.headers, json=data
        )
        self.log_function(f"HTTP {self.method} request send to `{self.url}`.")



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import abc
import os
import tempfile
from multiprocessing import Lock
from pprint import pformat
from typing import Any, Dict, List, Optional, Union

import attr
import cv2
import numpy as np

from geti_sdk.http_session import GetiSession

from .enums import MediaType
from .media_identifiers import (
    ImageIdentifier,
    MediaIdentifier,
    VideoFrameIdentifier,
    VideoIdentifier,
)
from .task_annotation_state import TaskAnnotationState
from .utils import (
    attr_value_serializer,
    numpy_from_buffer,
    str_to_datetime,
    str_to_media_type,
)


@attr.define
class MediaInformation:
    """
    Basic information about a media item in Intel® Geti™.

    :var display_url: URL that can be used to download the full size media entity
    :var height: Height of the media entity, in pixels
    :var width: Width of the media entity, in pixels
    :var size: Size of the media entity, in bytes
    """

    display_url: str
    height: int
    width: int
    extension: Optional[str] = None  # Added in Geti v2.5
    size: Optional[int] = None  # Added in Geti v1.2


@attr.define
class VideoInformation(MediaInformation):
    """
    Basic information about a video entity in Intel® Geti™.

    :var duration: Duration of the video
    :var frame_count: Total number of frames in the video
    :var frame_stride: Frame stride of the video
    :var frame_rate: Frame rate of the video
    """

    duration: int = attr.ib(kw_only=True)
    frame_count: int = attr.ib(kw_only=True)
    frame_stride: int = attr.ib(kw_only=True)
    frame_rate: Optional[float] = attr.ib(
        kw_only=True, default=None
    )  # Added in Geti v1.1


@attr.define
class ImageInformation(MediaInformation):
    """
    Basic information about an image entity in Intel® Geti™
    """

    pass


@attr.define
class VideoFrameInformation(MediaInformation):
    """
    Basic information about a video frame in Intel® Geti™.
    """

    frame_index: int = attr.ib(kw_only=True)
    video_id: str = attr.ib(kw_only=True)


@attr.define
class MediaItem:
    """
    Representation of a media entity in Intel® Geti™.

    :var id: Unique database ID of the media entity
    :var name: Filename of the media entity
    :var type: MediaType of the entity
    :var upload_time: Time and date at which the entity was uploaded to the system
    :var thumbnail: URL that can be used to get a thumbnail for the media entity
    :var annotation_state_per_task: Annotation state of the media entity
    :var media_information: Container holding basic information such as width and
        height about the media entity
    :param last_annotator_id: the name or id of the editor.
    """

    id: str
    name: str
    type: str = attr.field(converter=str_to_media_type)
    upload_time: str = attr.field(converter=str_to_datetime)
    media_information: MediaInformation
    annotation_state_per_task: Optional[List[TaskAnnotationState]] = None
    thumbnail: Optional[str] = None
    uploader_id: Optional[str] = None
    # last_annotator_id was added in Geti v 2.0. It replaced `editor_name`
    last_annotator_id: Optional[str] = None

    @property
    def download_url(self) -> str:
        """
        Return the URL that can be used to download the full size media entity from the
        Intel® Geti™ server.

        :return: URL at which the media entity can be downloaded
        """
        return self.media_information.display_url.strip()

    @property
    def base_url(self) -> str:
        """
        Return the base URL for the media item, which is the URL pointing to the
        media details of the entity.

        :return: Base URL of the media entity
        """
        display_url_image = "/display/full"
        display_url_video = "/display/stream"
        if self.download_url.endswith(display_url_image):
            url = self.download_url[: -len(display_url_image)]
        elif self.download_url.endswith(display_url_video):
            url = self.download_url[: -len(display_url_video)]
        elif self.download_url.endswith(self.id):
            url = self.download_url
        else:
            raise ValueError(
                f"Unexpected end pattern found in display URL for media entity {self}. "
                f"Unable to construct base_url for entity"
            )
        return url

    @property
    def identifier(self) -> MediaIdentifier:
        """
        Return the media identifier for the media item.

        :return: MediaIdentifier which uniquely identifies the media item.
        """
        return MediaIdentifier(type=self.type)

    @abc.abstractmethod
    def get_data(self, session: GetiSession) -> Union[np.ndarray, str]:
        """
        Get the pixel data for this MediaItem. Uses caching.

        For Image and VideoFrames, this method will return a np.ndarray. For Video
        objects, this method will return a path-like string, pointing to the video
        on the local disk.

        :param session: REST session to the Intel® Geti™ server on which the MediaItem
            lives
        :raises ValueError: If the cache is empty and no data can be downloaded
            from the cluster
        :return: numpy array holding the pixel data for the MediaItem, or string
            pointing to the file location for Videos.
        """
        raise NotImplementedError

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the MediaItem to a dictionary representation.

        :return: Dictionary holding the annotation scene data
        """
        output_dict = attr.asdict(self, value_serializer=attr_value_serializer)
        return output_dict

    @property
    def overview(self) -> str:
        """
        Return a string holding an overview of the media item.

        :return: overview string of the media item
        """
        return pformat(self.to_dict())


@attr.define(slots=False)
class Image(MediaItem):
    """
    Representation of an image in Intel® Geti™.

    :var media_information: Container holding basic information such as width and
            height about the image entity
    """

    media_information: ImageInformation = attr.field(kw_only=True)
    annotation_scene_id: Optional[str] = attr.field(
        kw_only=True, default=None, repr=False
    )  # Added in Geti v 1.16
    roi_id: Optional[str] = attr.field(
        kw_only=True, default=None, repr=False
    )  # Added in Geti v 1.16

    def __attrs_post_init__(self):
        """
        Initialize private attributes.
        """
        self._data: Optional[np.ndarray] = None

    @property
    def identifier(self) -> ImageIdentifier:
        """
        Return the media identifier for the Image instance

        :return: ImageIdentifier object that contains the identifiers of the image
        """
        return ImageIdentifier(image_id=self.id, type=self.type)

    def get_data(self, session: GetiSession) -> np.ndarray:
        """
        Get the pixel data for this Image. This method uses caching: If the cache is
        empty, it will download the data using the provided session. Otherwise it
        will return the cached data directly

        NOTE: The pixel data will be returned in BGR channel order

        :param session: REST session to the Intel® Geti™ server on which the Image lives
        :raises ValueError: If the cache is empty and no data can be downloaded
            from the cluster
        :return: Numpy.ndarray holding the pixel data for this Image.
        """
        if self._data is None:
            response = session.get_rest_response(
                url=self.download_url, method="GET", contenttype="jpeg"
            )
            if response.status_code == 200:
                self._data = numpy_from_buffer(response.content)
            else:
                raise ValueError(
                    f"Unable to retrieve data for image {self}, received "
                    f"response {response} from Intel® Geti™ server."
                )
        return self._data

    @property
    def numpy(self) -> Optional[np.ndarray]:
        """
        Pixel data for the Image, as a numpy array of shape (width x heigth x 3).
        If this attribute is None, the pixel data should be downloaded from the
        cluster first using the `get_data` method

        :return: numpy.ndarray containing the pixel data
        """
        return self._data


@attr.define(slots=False)
class VideoAnnotationStatistics:
    """
    Annotation statistics for a video in Intel® Geti™.

    :var annotated: Number of frames in the video that have been fully annotated
    :var partially_annotated: Number of frames in the video that have been partially
        annotated, meaning that one task in the task chain is still missing annotations
    :var unannotated: Number of frames in the video that have not yet been annotated
    """

    annotated: int
    partially_annotated: int
    unannotated: int


@attr.define(slots=False)
class Video(MediaItem):
    """
    Representation of a video in Intel® Geti™.

    :var media_information: Container holding basic information such as width,
            height and duration about the video entity
    """

    media_information: VideoInformation = attr.field(kw_only=True)
    matched_frames: Optional[int] = attr.field(kw_only=True, default=None, repr=False)
    annotation_statistics: Optional[VideoAnnotationStatistics] = attr.field(
        kw_only=True, default=None, repr=False
    )

    def __attrs_post_init__(self):
        """
        Initialize private attributes.
        """
        self._data: Optional[str] = None
        self._needs_tempfile_deletion: bool = False
        self._data_lock: Lock = Lock()

    @property
    def identifier(self) -> VideoIdentifier:
        """
        Return the media identifier for the Video instance.

        :return: VideoIdentifier object that contains the identifiers of the video
        """
        return VideoIdentifier(video_id=self.id, type=self.type)

    def get_data(self, session: GetiSession) -> str:
        """
        Get the video data from the Intel® Geti™ server. Calling this method will
        download the video data to a temporary file, and returns the path to file.

        :param session: GetiSession object pointing to the instance from which the video
            should be downloaded.
        :return: Path to the temporary video file on local disk.
        """
        if self._data is None:
            response = session.get_rest_response(
                url=self.download_url, method="GET", contenttype="jpeg"
            )
            if response.status_code == 200:
                file = tempfile.NamedTemporaryFile(delete=False, suffix=".mp4")
                file.write(response.content)
                file.close()
                self._data = file.name
                self._needs_tempfile_deletion = True
            else:
                raise ValueError(
                    f"Unable to retrieve data for image {self}, received "
                    f"response {response} from Intel® Geti™ server."
                )
        return self._data

    def to_frames(
        self, frame_stride: Optional[int] = None, include_data: bool = False
    ) -> List["VideoFrame"]:
        """
        Extract VideoFrames from the Video. Returns a list of VideoFrame objects.

        :param frame_stride: Stride to use for frame extraction. If left as None (the
            default), the frame_stride stored in the media_information is used.
        :param include_data: True to include pixel data for the frames, if available.
        :return: List of VideoFrames constructed from the Video
        """
        if frame_stride is None:
            frame_stride = self.media_information.frame_stride
        if frame_stride < 0:
            raise ValueError(
                f"Invalid frame stride {frame_stride}. Frame stride cannot be "
                f"negative."
            )
        elif frame_stride > self.media_information.frame_count:
            frame_stride = self.media_information.frame_count - 1
        frames_range = range(0, self.media_information.frame_count, frame_stride)
        if not include_data:
            frames = [
                VideoFrame.from_video(self, frame_index=index) for index in frames_range
            ]
        else:
            frames = [self.get_frame(frame_index=index) for index in frames_range]
        return frames

    def get_frame(self, frame_index: int) -> "VideoFrame":
        """
        Return a VideoFrame extracted from the Video, at the specified index.

        This method loads the numpy data for the frame, if available.

        :param frame_index: Index of the frame to extract
        :return: VideoFrame at the specified index
        """
        video_frame = VideoFrame.from_video(video=self, frame_index=frame_index)
        if self._data is not None:
            video_capture = cv2.VideoCapture(self._data)
            video_capture.set(cv2.CAP_PROP_POS_FRAMES, frame_index)
            success, frame = video_capture.read()
            if success:
                video_frame._data = frame
        return video_frame

    def __del__(self):
        """
        Clean up the temporary file created to store the video data (if any). This
        method is called when the Video object is deleted.
        """
        with self._data_lock:
            if self._needs_tempfile_deletion:
                if os.path.exists(self._data) and os.path.isfile(self._data):
                    os.remove(self._data)


@attr.define(slots=False)
class VideoFrame(MediaItem):
    """
    Representation of a video frame in Intel® Geti™.

    :var media_information: Container holding basic information such as width and
            height about the VideoFrame entity
    :var data: Pixel data for the VideoFrame. If this is None, the data can be
        downloaded using the 'get_data' method
    """

    media_information: VideoFrameInformation = attr.field(kw_only=True)
    video_name: Optional[str] = attr.field(kw_only=True, default=None)

    def __attrs_post_init__(self):
        """
        Initialize private attributes.
        """
        self._data: Optional[np.ndarray] = None

    @classmethod
    def from_video(cls, video: Video, frame_index: int) -> "VideoFrame":
        """
        Create a VideoFrame entity from a `video` and a `frame_index`.

        :param video: Video to extract the VideoFrame from
        :param frame_index: index at which the frame lives in the video
        :return:
        """
        base_url = f"{video.base_url}/frames/{frame_index}"
        frame_information = VideoFrameInformation(
            frame_index=frame_index,
            width=video.media_information.width,
            height=video.media_information.height,
            video_id=video.id,
            display_url=f"{base_url}/display/full",
        )
        return VideoFrame(
            name=f"{video.name}_frame_{frame_index}",
            type=str(MediaType.VIDEO_FRAME),
            upload_time=video.upload_time,
            thumbnail=f"{base_url}/display/thumb",
            media_information=frame_information,
            annotation_state_per_task=video.annotation_state_per_task,
            id=video.id,
            video_name=video.name,
        )

    @property
    def identifier(self) -> VideoFrameIdentifier:
        """
        Return the media identifier for the VideoFrame instance.

        :return: VideoFrameIdentifier object that contains the identifiers of the
            video frame
        """
        return VideoFrameIdentifier(
            video_id=self.id,
            type=self.type,
            frame_index=self.media_information.frame_index,
        )

    @property
    def numpy(self) -> Optional[np.ndarray]:
        """
        Pixel data for the Image, as a numpy array of shape (width x heigth x 3).
        If this attribute is None, the pixel data should be downloaded from the
        cluster first using the `get_data` method

        :return: numpy.ndarray containing the pixel data
        """
        return self._data

    def get_data(self, session: GetiSession) -> np.ndarray:
        """
        Get the pixel data for this VideoFrame. This method uses caching: If the cache
        is empty, it will download the data using the provided session. Otherwise it
        will return the cached data directly

        :param session: REST session to the Intel® Geti™ server on which the
            VideoFrame lives
        :raises ValueError: If the cache is empty and no data can be downloaded
            from the cluster
        :return: Numpy.ndarray holding the pixel data for this VideoFrame.
        """
        if self._data is None:
            response = session.get_rest_response(
                url=self.download_url, method="GET", contenttype="jpeg"
            )
            if response.status_code == 200:
                self._data = numpy_from_buffer(response.content)
            else:
                raise ValueError(
                    f"Unable to retrieve data for image {self}, received "
                    f"response {response} from Intel® Geti™ server."
                )
        return self._data



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import time
from contextlib import nullcontext
from typing import Any, Dict, List, Optional, Sequence, Union

from vcr import VCR

from geti_sdk import Geti
from geti_sdk.annotation_readers import AnnotationReader, DatumAnnotationReader
from geti_sdk.data_models import Project, TaskType
from geti_sdk.rest_clients import (
    AnnotationClient,
    ConfigurationClient,
    DatasetClient,
    ImageClient,
    ModelClient,
    PredictionClient,
    ProjectClient,
    TrainingClient,
    VideoClient,
)

from .constants import CASSETTE_EXTENSION, PROJECT_PREFIX
from .finalizers import force_delete_project


class ProjectService:
    """
    This class contains functionality to quickly create projects and interact with
    them through the respective clients

    :param geti: Geti instance representing the GETi server and workspace in
        which to create the project
    :param vcr: VCR instance used for recording HTTP requests made during the project
        lifespan. If left as None, no requests will be recorded or played back from
        VCR cassettes
    """

    def __init__(self, geti: Geti, vcr: Optional[VCR] = None, is_offline: bool = False):
        if vcr is None:
            self.vcr_context = nullcontext
        else:
            self.vcr_context = vcr.use_cassette
        self.session = geti.session
        self.workspace_id = geti.workspace_id
        self.geti = geti
        self.project_client = ProjectClient(
            session=geti.session, workspace_id=geti.workspace_id
        )

        self._project: Optional[Project] = None
        self._project_creation_timestamp: Optional[float] = None
        self._is_offline: bool = is_offline
        self._configuration_client: Optional[ConfigurationClient] = None
        self._image_client: Optional[ImageClient] = None
        self._annotation_client: Optional[AnnotationClient] = None
        self._training_client: Optional[TrainingClient] = None
        self._video_client: Optional[VideoClient] = None
        self._model_client: Optional[ModelClient] = None
        self._prediction_client: Optional[PredictionClient] = None
        self._dataset_client: Optional[DatasetClient] = None
        self._client_names = [
            "_configuration_client",
            "_image_client",
            "_annotation_client",
            "_training_client",
            "_video_client",
            "_model_client",
            "_prediction_client",
            "_dataset_client",
        ]
        self._project_removal_delay = 5  # seconds

    def create_project(
        self,
        project_name: Optional[str] = None,
        project_type: str = "classification",
        labels: Optional[List[Union[List[str], List[Dict[str, Any]]]]] = None,
    ) -> Project:
        """
        Create a project according to the `name`, `project_type` and `labels` specified.

        :param project_name: Name of the project to create
        :param project_type: Type of the project to create
        :param labels: List of labels for each task
        :return: the created project
        """
        if project_name is None:
            project_name = f"{PROJECT_PREFIX}_project_simple"
        if self._project is None:
            if labels is None:
                labels = [["cube", "cylinder"]]
            with self.vcr_context(f"{project_name}.{CASSETTE_EXTENSION}"):
                project = self.project_client.create_project(
                    project_name=project_name, project_type=project_type, labels=labels
                )
                self.project = project
                return project
        else:
            raise ValueError(
                "This ProjectService instance already contains an existing project. "
                "Please either delete the existing project first or use a new "
                "instance to create another project"
            )

    def get_or_create_project(
        self,
        project_name: str = "sdk_test_project_simple",
        project_type: str = "classification",
        labels: Optional[List[Union[List[str], List[Dict[str, Any]]]]] = None,
    ) -> Project:
        """
        This method will always return a project. It will either create a new one, or
        return the existing project if it has already been created.

        :param project_name: Name of the project to create
        :param project_type: Type of the project to create
        :param labels: List of labels for each task
        :return: the existing or newly created project
        """
        if not self.has_project:
            self.create_project(
                project_name=project_name, project_type=project_type, labels=labels
            )
        return self.project

    def create_project_from_dataset(
        self,
        annotation_readers: List[AnnotationReader],
        project_name: str = "sdk_test_project_simple",
        project_type: str = "classification",
        path_to_dataset: str = "",
        n_images: int = 12,
    ):
        """
        Create a project from a dataset, and upload media and annotations to it

        :param annotation_readers: List of annotation readers that should be used as
            annotation sources for the tasks in the project
        :param project_name: Name of the project to create
        :param project_type: Type of the project to create
        :param path_to_dataset: Path to the base folder containing the dataset
        :param n_images: Number of annotated images to upload. If set to -1, uploads
            all images for which the annotation reader contains annotations
        :raises: ValueError in case the ProjectService already contains a project
        :return: The Project that was created
        """
        if not self.has_project:
            with self.vcr_context(f"{project_name}.{CASSETTE_EXTENSION}"):
                if len(annotation_readers) == 1:
                    annotation_reader = annotation_readers[0]
                    project = self.geti.create_single_task_project_from_dataset(
                        project_name=project_name,
                        project_type=project_type,
                        path_to_images=path_to_dataset,
                        annotation_reader=annotation_reader,
                        number_of_images_to_upload=n_images,
                        number_of_images_to_annotate=n_images,
                        enable_auto_train=False,
                        upload_videos=True,
                        max_threads=1,
                    )
                else:
                    project = self.geti.create_task_chain_project_from_dataset(
                        project_name=project_name,
                        project_type=project_type,
                        path_to_images=path_to_dataset,
                        label_source_per_task=annotation_readers,
                        number_of_images_to_upload=n_images,
                        number_of_images_to_annotate=n_images,
                        enable_auto_train=False,
                        max_threads=1,
                    )
        else:
            raise ValueError(
                "This ProjectService instance already contains an existing project. "
                "Please either delete the existing project first or use a new "
                "instance to create another project"
            )
        self.project = project
        return project

    @property
    def project(self) -> Project:
        """
        Returns the project managed by the ProjectService.

        :return: The project managed by the ProjectService
        :raises: ValueError if the ProjectService does not contain a project yet
        """
        if self._project is None:
            raise ValueError(
                "This ProjectService instance does not contain a project yet. Please "
                "call `ProjectService.create_project` to create a new project first."
            )
        return self._project

    @project.setter
    def project(self, value: Optional[Project]) -> None:
        """
        Set the project for the ProjectService.
        """
        self._project = value
        if self._project is None:
            self._project_creation_timestamp = None
        else:
            self._project_creation_timestamp = time.time()

    @property
    def has_project(self) -> bool:
        """
        Returns True if the ProjectService contains an existing project, False otherwise
        """
        return self._project is not None

    @property
    def is_task_chain_project(self) -> bool:
        """
        Returns True if the project belonging to this ProjectService instance is a task
        chain project, False otherwise

        If this method is called while no project is defined yet, a ValueError will be
        raised
        """
        return len(self.project.get_trainable_tasks()) > 1

    @property
    def image_client(self) -> ImageClient:
        """Returns the ImageClient instance for the project"""
        if self._image_client is None:
            with self.vcr_context(
                f"{self.project.name}_image_client.{CASSETTE_EXTENSION}"
            ):
                self._image_client = ImageClient(
                    session=self.session,
                    workspace_id=self.workspace_id,
                    project=self.project,
                )
        return self._image_client

    @property
    def video_client(self) -> VideoClient:
        """Returns the VideoClient instance for the project"""
        if self._video_client is None:
            with self.vcr_context(
                f"{self.project.name}_video_client.{CASSETTE_EXTENSION}"
            ):
                self._video_client = VideoClient(
                    session=self.session,
                    workspace_id=self.workspace_id,
                    project=self.project,
                )
        return self._video_client

    @property
    def annotation_client(self) -> AnnotationClient:
        """Returns the AnnotationClient instance for the project"""
        if self._annotation_client is None:
            with self.vcr_context(
                f"{self.project.name}_annotation_client.{CASSETTE_EXTENSION}"
            ):
                self._annotation_client = AnnotationClient(
                    session=self.session,
                    workspace_id=self.workspace_id,
                    project=self.project,
                )
        return self._annotation_client

    @property
    def dataset_client(self) -> DatasetClient:
        """Returns the DatasetClient instance for the project"""
        if self._dataset_client is None:
            with self.vcr_context(
                f"{self.project.name}_dataset_client.{CASSETTE_EXTENSION}"
            ):
                self._dataset_client = DatasetClient(
                    session=self.session,
                    workspace_id=self.workspace_id,
                    project=self.project,
                )
        return self._dataset_client

    @property
    def configuration_client(self) -> ConfigurationClient:
        """Returns the ConfigurationClient instance for the project"""
        if self._configuration_client is None:
            with self.vcr_context(
                f"{self.project.name}_configuration_client.{CASSETTE_EXTENSION}"
            ):
                self._configuration_client = ConfigurationClient(
                    session=self.session,
                    workspace_id=self.workspace_id,
                    project=self.project,
                )
        return self._configuration_client

    @property
    def training_client(self) -> TrainingClient:
        """Returns the TrainingClient instance for the project"""
        if self._training_client is None:
            with self.vcr_context(
                f"{self.project.name}_training_client.{CASSETTE_EXTENSION}"
            ):
                self._training_client = TrainingClient(
                    session=self.session,
                    workspace_id=self.workspace_id,
                    project=self.project,
                )
        return self._training_client

    @property
    def prediction_client(self) -> PredictionClient:
        """Returns the PredictionClient instance for the project"""
        if self._prediction_client is None:
            with self.vcr_context(
                f"{self.project.name}_prediction_client.{CASSETTE_EXTENSION}"
            ):
                self._prediction_client = PredictionClient(
                    session=self.session,
                    workspace_id=self.workspace_id,
                    project=self.project,
                )
        return self._prediction_client

    @property
    def model_client(self) -> ModelClient:
        """Returns the ModelClient instance for the project"""
        if self._model_client is None:
            with self.vcr_context(
                f"{self.project.name}_model_client.{CASSETTE_EXTENSION}"
            ):
                self._model_client = ModelClient(
                    session=self.session,
                    workspace_id=self.workspace_id,
                    project=self.project,
                )
        return self._model_client

    @property
    def is_training(self) -> bool:
        """
        Returns True if the project service contains a project, and that project is
        currently running at least one training job for any of its tasks. Returns
        False otherwise.

        :return: True if the project is training, False otherwise
        """
        if not self.has_project:
            return False
        with self.vcr_context(f"{self.project.name}_is_training.{CASSETTE_EXTENSION}"):
            return self.training_client.is_training()

    def delete_project(self):
        """Deletes the project from the server"""
        if self.has_project:
            with self.vcr_context(f"{self.project.name}_deletion.{CASSETTE_EXTENSION}"):
                # Server needs a moment to process the project before deletion
                if (
                    not self._is_offline
                    and (lifetime := time.time() - self._project_creation_timestamp)
                    < self._project_removal_delay
                ):
                    time.sleep(self._project_removal_delay - lifetime)
                force_delete_project(self.project, self.project_client)
                self.reset_state()

    def reset_state(self) -> None:
        """
        Resets the state of the ProjectService instance. This method should be called
        once the project belonging to the project service is deleted from the server
        """
        self.project = None
        for client_name in self._client_names:
            setattr(self, client_name, None)

    def add_annotated_media(
        self, annotation_readers: Sequence[AnnotationReader], n_images: int = 12
    ) -> None:
        """
        Adds annotated media to the project

        :param annotation_readers: List of AnnotationReader instances to use to get
            annotations for the media. The length of this list must match the number
            of trainable tasks in the project
        :param n_images: Number of annotated images to upload. If set to -1, uploads
            all images for which the annotation reader contains annotations
        """
        annotation_reader_1 = annotation_readers[0]
        data_path = annotation_reader_1.base_folder

        if len(annotation_readers) != len(self.project.get_trainable_tasks()):
            raise ValueError(
                "Number of annotation readers received does not match the number of "
                "trainable tasks in the project: Unable to upload annotated media"
            )
        with self.vcr_context(
            f"{self.project.name}_add_annotated_media.{CASSETTE_EXTENSION}"
        ):
            if isinstance(annotation_reader_1, DatumAnnotationReader):
                images = self.image_client.upload_from_list(
                    path_to_folder=data_path,
                    image_names=annotation_reader_1.get_all_image_names(),
                    n_images=n_images,
                    max_threads=1,
                )
            else:
                images = self.image_client.upload_folder(
                    data_path, n_images=n_images, max_threads=1
                )

            if n_images < len(images) and n_images != -1:
                images = images[:n_images]
            # Annotation preparation and upload
            for task_index, task in enumerate(self.project.get_trainable_tasks()):
                # Set annotation reader task type
                annotation_readers[task_index].task_type = task.type
                annotation_readers[task_index].prepare_and_set_dataset(
                    task_type=task.type
                )
                self.annotation_client.annotation_reader = annotation_readers[
                    task_index
                ]
                # Upload annotations
                self.annotation_client.upload_annotations_for_images(
                    images=images, append_annotations=task_index > 0, max_threads=1
                )

    def set_auto_train(self, auto_train: bool = True) -> None:
        """
        Sets the 'auto_training' parameter for all tasks in the project to `auto_train`

        :param auto_train: True to turn auto_training on, False to turn it off
        """
        with self.vcr_context(
            f"{self.project.name}_set_auto_train.{CASSETTE_EXTENSION}"
        ):
            self.configuration_client.set_project_auto_train(auto_train=auto_train)

    def set_minimal_training_hypers(self) -> None:
        """
        Configures the project to use the lowest possible batch size and number of
        epochs to perform a minimal training round

        """
        with self.vcr_context(
            f"{self.project.name}_set_minimal_hypers.{CASSETTE_EXTENSION}"
        ):
            self.configuration_client.set_project_num_iterations(1)
            for task in self.project.get_trainable_tasks():
                task_config = self.configuration_client.get_task_configuration(task.id)
                try:
                    task_config.set_parameter_value("batch_size", 1)
                    self.configuration_client.set_configuration(task_config)
                except ValueError:
                    logging.warning(
                        f"Parameter batch_size was not found in the configuration for "
                        f"task {task.summary}. Unable to configure batch size"
                    )

    def set_reduced_memory_hypers(self) -> None:
        """
        Reduce batch size in memory intensive tasks to avoid OOM errors in pods. Use
        default hypers for other tasks
        """
        with self.vcr_context(
            f"{self.project.name}_set_reduced_memory_hypers.{CASSETTE_EXTENSION}"
        ):
            for task in self.project.get_trainable_tasks():
                if task.type in [
                    TaskType.DETECTION,
                    TaskType.ROTATED_DETECTION,
                    TaskType.INSTANCE_SEGMENTATION,
                ]:
                    task_hypers = self.configuration_client.get_task_configuration(
                        task_id=task.id
                    )
                    task_hypers.batch_size.value = 1
                    self.configuration_client.set_configuration(task_hypers)

    def set_auto_training_annotation_requirement(
        self, required_images: int = 6
    ) -> None:
        """
        Sets the 'Number of images required for auto-training' parameter for all
        tasks in the project to `required_images`

        :param required_images: Number of images required before starting a new round
            of auto training for the task
        """
        with self.vcr_context(
            f"{self.project.name}_set_auto_training_annotation_requirement.{CASSETTE_EXTENSION}"
        ):
            self.configuration_client.set_project_parameter(
                parameter_name="required_images_auto_training", value=required_images
            )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import os
import time
import warnings
from concurrent.futures import ThreadPoolExecutor
from glob import glob
from typing import (
    Any,
    BinaryIO,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    Type,
)

from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from geti_sdk.data_models import Dataset, Image, MediaType, Project, Video, VideoFrame
from geti_sdk.data_models.containers.media_list import MediaList, MediaTypeVar
from geti_sdk.data_models.enums.media_type import (
    SUPPORTED_IMAGE_FORMATS,
    SUPPORTED_VIDEO_FORMATS,
)
from geti_sdk.data_models.utils import numpy_from_buffer
from geti_sdk.http_session import GetiRequestException, GetiSession
from geti_sdk.rest_clients.dataset_client import DatasetClient
from geti_sdk.rest_converters.media_rest_converter import MediaRESTConverter

MEDIA_TYPE_MAPPING = {MediaType.IMAGE: Image, MediaType.VIDEO: Video}
MEDIA_SUPPORTED_FORMAT_MAPPING = {
    MediaType.IMAGE: SUPPORTED_IMAGE_FORMATS,
    MediaType.VIDEO: SUPPORTED_VIDEO_FORMATS,
}
MEDIA_DOWNLOAD_FORMAT_MAPPING = {MediaType.IMAGE: ".jpg", MediaType.VIDEO: ".mp4"}


class BaseMediaClient(Generic[MediaTypeVar]):
    """
    Class to manage media up and downloads for a certain project.
    """

    _MEDIA_TYPE: ClassVar[MediaType]

    def __init__(self, session: GetiSession, workspace_id: str, project: Project):
        self.session = session
        self._workspace_id = workspace_id
        self._base_url = f"workspaces/{workspace_id}/projects/{project.id}/datasets"
        self._project = project
        self.__media_type: Type[MediaTypeVar] = self.__get_media_type(self._MEDIA_TYPE)
        self._dataset_client = DatasetClient(
            session=session, project=project, workspace_id=workspace_id
        )

    def base_url(self, dataset: Dataset) -> str:
        """
        Return the base url for the media endpoint.

        :param dataset: Dataset to retrieve the base url for
        :return: string containing the base url
        """
        return f"{self._base_url}/{dataset.id}/media/{self.plural_media_name}"

    @property
    def plural_media_name(self) -> str:
        """
        Convert the type of the media entities managed by this media client to a
        string in plural form.

        :return:
        """
        return f"{self._MEDIA_TYPE}s"

    @staticmethod
    def __get_media_type(media_type: MediaType) -> Type[MediaTypeVar]:
        """
        Get the type of the media entities that are managed by this media client.

        :param media_type: MediaType Enum instance representing the type of media
            entities
        :return: Type of media entities that can be used to instantiate objects of this
            type
        """
        return MEDIA_TYPE_MAPPING[media_type]

    def _get_all(self, dataset: Optional[Dataset] = None) -> MediaList[MediaTypeVar]:
        """
        Get a list holding all media entities of a certain type in the project.

        :param dataset: Optional dataset to retrieve the media for. If no dataset is
            specified, only media from the training dataset will be returned
        :return: MediaList holding all media of a certain type in the project
        """
        if dataset is None:
            dataset = self._project.training_dataset

        url = f"{self._base_url}/{dataset.id}/media:query?limit=100"
        data = {
            "condition": "and",
            "rules": [
                {
                    "field": "MEDIA_TYPE",
                    "operator": "EQUAL",
                    "value": f"{self._MEDIA_TYPE}",
                }
            ],
        }
        response = self.session.get_rest_response(url=url, method="POST", data=data)
        total_number_of_media: int = response[f"total_matched_{self.plural_media_name}"]

        raw_media_list: List[Dict[str, Any]] = []
        while len(raw_media_list) < total_number_of_media:
            for media_item_dict in response["media"]:
                raw_media_list.append(media_item_dict)
            if "next_page" in response.keys():
                response = self.session.get_rest_response(
                    url=response["next_page"],
                    method="POST",
                    data=data,
                    include_organization_id=False,
                )
        return MediaList.from_rest_list(
            rest_input=raw_media_list, media_type=self.__media_type
        )

    def _delete_media(self, media_list: Sequence[MediaTypeVar]) -> bool:
        """
        Delete all media entities in `media_list` from the project.

        :param media_list: List of media entities to delete
        :return: True if all items on the media list were deleted successfully,
            False otherwise
        """
        if not isinstance(media_list, MediaList) and isinstance(media_list, Sequence):
            media_list = MediaList(media_list)
        if media_list.media_type == VideoFrame:
            raise ValueError("Unable to delete individual video frames.")
        logging.info(
            f"Deleting {len(media_list)} {self.plural_media_name} from project "
            f"'{self._project.name}'..."
        )
        for media_item in media_list:
            try:
                self.session.get_rest_response(url=media_item.base_url, method="DELETE")
            except GetiRequestException as error:
                if error.status_code == 409:
                    logging.info(
                        f"Project '{self._project.name}' is locked for deletion, "
                        f"unable to delete media. Aborting deletion."
                    )
                    return False
                if error.status_code == 404 or error.status_code == 204:
                    # Media item has already been deleted, continue with the rest of
                    # the list
                    continue
        return True

    def _upload_bytes(
        self, buffer: BinaryIO, dataset: Optional[Dataset] = None
    ) -> Dict[str, Any]:
        """
        Upload a buffer representing a media file to the server.

        :param buffer: BinaryIO object representing a media file
        :param dataset: Dataset to upload the media to. If no dataset is passed, the
            media will be uploaded into the default (training) dataset
        :return: Dictionary containing the response of the Intel® Geti™ server, which
            holds the details of the uploaded entity
        """
        if dataset is None:
            dataset = self._project.training_dataset
        response = self.session.get_rest_response(
            url=f"{self.base_url(dataset)}",
            method="POST",
            contenttype="multipart",
            data={"file": buffer},
        )
        return response

    def _upload(
        self, filepath: str, dataset: Optional[Dataset] = None
    ) -> Dict[str, Any]:
        """
        Upload a media file to the server.

        :param filepath: full path to the media file on disk
        :param dataset: Dataset to upload the media to. If no dataset is passed, the
            media will be uploaded into the default (training) dataset
        :return: Dictionary containing the response of the Intel® Geti™ server, which
            holds the details of the uploaded entity
        """
        with open(filepath, "rb") as f:
            response = self._upload_bytes(f, dataset=dataset)
        return response

    def _upload_loop(
        self,
        filepaths: List[str],
        skip_if_filename_exists: bool = False,
        dataset: Optional[Dataset] = None,
        max_threads: int = 5,
    ) -> MediaList[MediaTypeVar]:
        """
        Upload media from a list of filepaths. Also checks if media items with the same
        filename exist in the project, to make sure no duplicate items are uploaded.

        :param filepaths: List of full filepaths for media that should be
            uploaded
        :param skip_if_filename_exists: Set to True to skip uploading of a media item
            if a media item with the same filename already exists in the dataset.
            Defaults to False
        :param dataset: Dataset to upload the media to. If no dataset is passed, the
            media will be uploaded into the default (training) dataset
        :param max_threads: Maximum number of threads to use for uploading. Defaults to 5.
            Set to -1 to use all available threads.
        :return: MediaList containing a list of all media entities that were uploaded
            to the project
        """
        if max_threads <= 0:
            # ThreadPoolExecutor will use minimum 5 threads for 1 core cpu
            # and maximum 32 threads for multi-core cpu.
            max_threads = None
        if dataset is None:
            dataset = self._project.training_dataset
        media_in_project = self._get_all(dataset=dataset)
        uploaded_media: MediaList[MediaTypeVar] = MediaList[MediaTypeVar]([])
        upload_count = 0
        skip_count = 0
        logging.info(
            f"Starting {self._MEDIA_TYPE} upload to dataset '{dataset.name}'..."
        )
        tqdm_prefix = f"Uploading {self.plural_media_name}"

        t_start = time.time()

        def upload_file(filepath: str) -> None:
            nonlocal upload_count, skip_count
            name, ext = os.path.splitext(os.path.basename(filepath))
            if name in media_in_project.names and skip_if_filename_exists:
                skip_count += 1
                return
            try:
                media_dict = self._upload(filepath=filepath, dataset=dataset)
            except GetiRequestException as error:
                if error.status_code == 500:
                    logging.error(
                        f"Failed to upload {self._MEDIA_TYPE} '{name}'. Error message: "
                        f"{error}"
                    )
                    return
                else:
                    raise error
            media_item = MediaRESTConverter.from_dict(
                input_dict=media_dict, media_type=self.__media_type
            )
            if isinstance(media_item, Video):
                media_item._data = filepath
            # appends are thread safe
            media_in_project.append(media_item)
            uploaded_media.append(media_item)
            upload_count += 1

        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            with logging_redirect_tqdm(tqdm_class=tqdm):
                list(
                    tqdm(  # List unwraps the generator
                        executor.map(upload_file, filepaths),
                        total=len(filepaths),
                        desc=tqdm_prefix,
                    )
                )

        t_elapsed = time.time() - t_start
        if upload_count > 0:
            msg = (
                f"Upload complete. Uploaded {upload_count} new "
                f"{self.plural_media_name} in {t_elapsed:.1f} seconds."
            )
        else:
            msg = f"No new {self.plural_media_name} were uploaded."
        if skip_count > 0:
            msg = (
                msg + f" Found {skip_count} {self.plural_media_name} that already "
                f"existed in project, these {self.plural_media_name} were"
                f" skipped."
            )
        logging.info(msg)
        return uploaded_media

    def _upload_folder(
        self,
        path_to_folder: str,
        n_media: int = -1,
        skip_if_filename_exists: bool = False,
        dataset: Optional[Dataset] = None,
        max_threads: int = 5,
    ) -> MediaList[MediaTypeVar]:
        """
        Upload all media in a folder to the project. Returns the mapping of filenames
        to the unique IDs assigned by Intel Geti.

        :param path_to_folder: Folder with media items to upload
        :param n_media: Number of media to upload from folder
        :param skip_if_filename_exists: Set to True to skip uploading of a media item
            if a media item with the same filename already exists in the project.
            Defaults to False
        :param dataset: Dataset to upload the media to. If no dataset is passed, the
            media will be uploaded into the default (training) dataset
        :param max_threads: Maximum number of threads to use for uploading. Defaults to 5.
            Set to -1 to use all available threads.
        :return: MediaList containing a list of all media entities that were uploaded
            to the project
        """
        media_formats = MEDIA_SUPPORTED_FORMAT_MAPPING[self._MEDIA_TYPE]
        filepaths: List[str] = []
        for media_extension in media_formats:
            filepaths += glob(
                os.path.join(path_to_folder, "**", f"*{media_extension}"),
                recursive=True,
            )
        n_files = len(filepaths)
        if n_media == -1:
            n_to_upload = n_files
        elif n_media < -1:
            raise ValueError(
                f"Number of {self.plural_media_name} to upload must positive, or -1 "
                f"to signify uploading all files in the folder. "
            )
        else:
            n_to_upload = n_files if n_files < n_media else n_media
        return self._upload_loop(
            filepaths=filepaths[0:n_to_upload],
            skip_if_filename_exists=skip_if_filename_exists,
            dataset=dataset,
            max_threads=max_threads,
        )

    def _download_all(
        self,
        path_to_folder: str,
        append_media_uid: bool = False,
        max_threads: int = 10,
        dataset: Optional[Dataset] = None,
    ) -> None:
        """
        Download all media entities in a project to a folder on the local disk.

        :param path_to_folder: path to the folder in which the media should be saved
        :param append_media_uid: True to append the UID of a media item to the
            filename (separated from the original filename by an underscore, i.e.
            '{filename}_{media_id}').
        :param max_threads: Maximum number of threads to use for downloading.
            Defaults to 10. Set to -1 to use all available threads.
        :return:
        """
        datasets = self._dataset_client.get_all_datasets()
        if dataset is not None:
            if dataset not in datasets:
                warnings.warn(
                    f"Dataset '{dataset.name}' not found in project '{self._project.name}'. "
                    f"Skipping download for this dataset."
                )
                return
            datasets = [dataset]
        for dataset in datasets:
            path_to_media_folder = os.path.join(
                path_to_folder, self.plural_media_name, dataset.name
            )
            self._download_dataset(
                dataset=dataset,
                path_to_media_folder=path_to_media_folder,
                append_media_uid=append_media_uid,
                max_threads=max_threads,
            )

    def _download_dataset(
        self,
        dataset: Dataset,
        path_to_media_folder: str,
        append_media_uid: bool = False,
        max_threads: int = 10,
    ):
        """
        Download all media items of a single type in the dataset to a folder on disk

        :param dataset: Dataset to download the media for
        :param path_to_media_folder: path to the folder in which the media should be
            saved
        :param append_media_uid: True to append the UID of a media item to the
            filename (separated from the original filename by an underscore, i.e.
            '{filename}_{media_id}').
        :param max_threads: Maximum number of threads to use for downloading.
            Defaults to 10. Set to -1 to use all available threads.
        """
        if max_threads <= 0:
            # ThreadPoolExecutor will use minimum 5 threads for 1 core cpu
            # and maximum 32 threads for multi-core cpu.
            max_threads = None
        media_list = self._get_all(dataset=dataset)
        os.makedirs(path_to_media_folder, exist_ok=True, mode=0o770)
        logging.info(
            f"Downloading {len(media_list)} {self.plural_media_name} from project "
            f"'{self._project.name}' and dataset '{dataset.name}' to folder "
            f"{path_to_media_folder}..."
        )
        t_start = time.time()
        download_count = 0
        existing_count = 0
        tqdm_prefix = f"Downloading {self.plural_media_name}"

        def download_file(media_item: MediaTypeVar) -> None:
            nonlocal download_count, existing_count
            uid_string = ""
            if append_media_uid:
                uid_string = f"_{media_item.id}"
            media_filepath = os.path.join(
                path_to_media_folder,
                os.path.basename(media_item.name)
                + uid_string
                + MEDIA_DOWNLOAD_FORMAT_MAPPING[self._MEDIA_TYPE],
            )
            if os.path.exists(media_filepath) and os.path.isfile(media_filepath):
                existing_count += 1
                return
            try:
                response = self.session.get_rest_response(
                    url=media_item.download_url,
                    method="GET",
                    contenttype="jpeg",
                    include_organization_id=False,
                )
            except GetiRequestException as error:
                if error.status_code == 500:
                    logging.error(
                        f"Failed to download {self._MEDIA_TYPE} '{media_item.name}' "
                        f"with ID '{media_item.id}'. Error message: {error}"
                    )
                    return
                else:
                    raise error

            with open(media_filepath, "wb") as f:
                f.write(response.content)
            if isinstance(media_item, (Image, VideoFrame)):
                # Set the numpy data attribute if the media item supports it
                media_item._data = numpy_from_buffer(response.content)
            elif isinstance(media_item, Video):
                media_item._data = media_filepath
            download_count += 1

        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            with logging_redirect_tqdm(tqdm_class=tqdm):
                list(
                    tqdm(  # List unwraps the generator
                        executor.map(download_file, media_list),
                        total=len(media_list),
                        desc=tqdm_prefix,
                    )
                )

        t_elapsed = time.time() - t_start
        if download_count > 0:
            msg = (
                f"Downloaded {download_count} {self.plural_media_name} in "
                f"{t_elapsed:.1f} seconds."
            )
        else:
            msg = f"No {self.plural_media_name} were downloaded."
        if existing_count > 0:
            msg += (
                f" {existing_count} existing {self.plural_media_name} were found "
                f"in the target folder, download was skipped for these "
                f"{self.plural_media_name}."
            )
        logging.info(msg)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
import json
import logging
import os
import time
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from geti_sdk.data_models import (
    AnnotationKind,
    Dataset,
    Image,
    MediaItem,
    Prediction,
    Project,
    Video,
    VideoFrame,
)
from geti_sdk.data_models.containers import MediaList
from geti_sdk.data_models.enums import MediaType, PredictionMode
from geti_sdk.data_models.predictions import ResultMedium
from geti_sdk.http_session import GetiRequestException, GetiSession
from geti_sdk.rest_converters.prediction_rest_converter import PredictionRESTConverter


class PredictionClient:
    """
    Class to download predictions from an existing Intel® Geti™ project.
    """

    def __init__(self, session: GetiSession, project: Project, workspace_id: str):
        self.session = session
        self.project = project
        self._base_url = f"workspaces/{workspace_id}/projects/{project.id}/"
        self._labels = project.get_all_labels()
        self.__project_ready = self.__are_models_trained()
        self._mode = PredictionMode.AUTO
        self.__override_mode: Optional[PredictionMode] = None

    def __are_models_trained(self) -> bool:
        """
        Check that the project to which this PredictionClient belongs has trained
        models to generate predictions from. This method will return True if at least
        one model is trained for each task in the project task chain.

        :return: True if the project has trained models and is ready to generate
            predictions, False otherwise
        """
        response = self.session.get_rest_response(
            url=f"{self._base_url}model_groups", method="GET"
        )
        model_info_array: List[Dict[str, Any]]

        model_info_array = response.get("model_groups", [])

        task_ids = [task.id for task in self.project.get_trainable_tasks()]
        tasks_with_models: List[str] = []
        for item in model_info_array:
            if len(item["models"]) > 0:
                tasks_with_models.append(item["task_id"])

        # It is sufficient if the FIRST task has a model trained
        if task_ids[0] in tasks_with_models:
            return True
        else:
            return False

    @property
    def ready_to_predict(self):
        """
        Return True if the project is ready to yield predictions, False otherwise.

        :return:
        """
        if self.__project_ready:
            return True
        else:
            self.__project_ready = self.__are_models_trained()
            return self.__project_ready

    @property
    def mode(self) -> PredictionMode:
        """
        Return the current mode used to retrieve predictions. There are three options:
         - auto
         - latest
         - online

        Auto will fetch prediction from the database if it is up to date and
        otherwise send an inference request. Online will always send an
        inference request and latest will not send an inference request but grabs
        the latest result from the database.

        By default, the mode is set to `auto`.

        :return: Current PredictionMode used to retrieve predictions
        """
        if self.__override_mode is None:
            return self._mode
        else:
            return self.__override_mode

    @mode.setter
    def mode(self, new_mode: Union[str, PredictionMode]):
        """
        Set the mode for the Prediction client to retrieve predictions from the
        Intel® Geti™ server.

        :param new_mode: PredictionMode (or string representing a prediction mode) to
            set
        """
        if isinstance(new_mode, str):
            new_mode = PredictionMode(new_mode)
        self._mode = new_mode

    def _get_prediction_for_media_item(
        self,
        media_item: MediaItem,
        dataset: Optional[Dataset] = None,
        prediction_mode: PredictionMode = PredictionMode.AUTO,
        include_explanation: bool = False,
    ) -> Tuple[Optional[Union[Prediction, List[Prediction]]], str]:
        """
        Get the prediction for a media item. If a 2D media item (Image or VideoFrame)
        is passed, this method will return a single Prediction. If a Video is passed,
        this method will return a list of predictions.

        In case of failure to get a prediction, the first element of the tuple
        returned by this method will be None, and the second will be a message
        describing the problem.

        :param media_item: Image, Video or VideoFrame to get the prediction for
        :param dataset: Dataset to which the media item belongs. Only needs to be
            specified when the project contains more than one dataset
        :param prediction_mode: Mode for the predictions to be retrieved. Use
            ONLINE to always generate a new prediction, or LATEST to obtain the latest
            prediction from the cache. AUTO can be used to first query the cache for
            a prediction, and if none is found the system will generate a new one on
            the fly
        :param include_explanation: Only applicable to Geti v1.13 or higher: True
            to include saliency maps in the result. If set to False, saliency maps
            will be omitted. Setting this to True will increase the time required to
            retrieve the prediction
        :return: Tuple containing:
         - Prediction (for Image/VideoFrame) or List of Predictions (for Video)
         - string containing a message
        """
        if not self.ready_to_predict:
            msg = (
                f"Not all tasks in project '{self.project.name}' have a trained "
                f"model available. Unable to get predictions from the project."
            )
            return None, msg
        if dataset is None:
            if len(self.project.datasets) == 1:
                dataset = self.project.training_dataset
                dataset_id = dataset.id
            else:
                logging.debug(
                    f"The project {self.project.name} contains multiple datasets, but "
                    f"the dataset for the media item was not specified. The "
                    f"PredictionClient will try to compute the correct dataset for the "
                    f"media item, but this may lead to a 404 `resource_not_found` error. "
                    f"For optimal performance, please make sure to specify the correct "
                    f"dataset"
                )
                dataset_id_string = media_item.base_url.split("/datasets/")[-1]
                dataset_id = dataset_id_string.split("/")[0]
        else:
            dataset_id = dataset.id
        data: Optional[Dict[str, str]] = None
        explain_response: Optional[dict] = None

        prediction_mode_map = {
            "auto": "auto",
            "online": "never",
            "latest": "always",
        }
        if media_item.type == MediaType.IMAGE:
            data = {
                "image_id": media_item.identifier.image_id,
                "dataset_id": dataset_id,
            }
            action = "predict"
        elif media_item.type == MediaType.VIDEO_FRAME:
            data = {
                "dataset_id": dataset_id,
                "video_id": media_item.identifier.video_id,
                "frame_index": str(media_item.identifier.frame_index),
            }
            action = "predict"
        elif media_item.type == MediaType.VIDEO:
            stride = media_item.media_information.frame_stride
            data = {
                "dataset_id": dataset_id,
                "video_id": media_item.identifier.video_id,
                "start_frame": 0,
                "end_frame": int(15 * stride),
                "frame_skip": stride,
            }
            action = "batch_predict"
        url = (
            f"{self._base_url}pipelines/active:{action}"
            f"?use_cache={prediction_mode_map[str(prediction_mode)]}"
        )
        include_org_id = True
        method = "POST"
        if include_explanation:
            explain_map = {"predict": "explain", "batch_predict": "batch_explain"}
            explain_url = f"{self._base_url}pipelines/active:{explain_map[action]}"
            explain_response = self.session.get_rest_response(
                url=explain_url,
                method="POST",
                include_organization_id=True,
                data=data,
            )
        try:
            response = self.session.get_rest_response(
                url=url,
                method=method,
                include_organization_id=include_org_id,
                data=data,
            )
            if isinstance(media_item, (Image, VideoFrame)):
                result = PredictionRESTConverter.from_dict(response)
                if include_explanation:
                    maps: List[ResultMedium] = []
                    for map_dict in explain_response.get("maps", []):
                        map = ResultMedium(
                            name="saliency_map", label_id=map_dict.get("label_id", None)
                        )
                        map.data = map_dict["data"].encode("utf-8")
                        maps.append(map)
                    result.maps = maps
                result.resolve_labels_for_result_media(labels=self._labels)
                result.resolve_label_names_and_colors(labels=self._labels)

            elif isinstance(media_item, Video):
                result = []
                for ind, prediction in enumerate(response["video_predictions"]):
                    pred_object = PredictionRESTConverter.from_dict(prediction)
                    pred_object.resolve_label_names_and_colors(labels=self._labels)
                    if include_explanation:
                        maps: List[ResultMedium] = []
                        for map_dict in explain_response["explanations"][ind].get(
                            "maps", []
                        ):
                            map = ResultMedium(
                                name="saliency_map",
                                label_id=map_dict.get("label_id", None),
                            )
                            map.data = map_dict["data"].encode("utf-8")
                            maps.append(map)
                        pred_object.maps = maps
                    pred_object.resolve_labels_for_result_media(labels=self._labels)
                    result.append(pred_object)
            else:
                raise TypeError(
                    f"Getting predictions is not supported for media item of type "
                    f"{media_item.type}. Unable to retrieve predictions."
                )
            msg = "success"
        except GetiRequestException as error:
            msg = f"Unable to retrieve prediction for {media_item.type}."
            if error.status_code == 204:
                msg += (
                    f" The prediction for the {media_item.type} with name "
                    f"'{media_item.name}' is not available in project "
                    f"'{self.project.name}'."
                )
                if prediction_mode == PredictionMode.LATEST:
                    msg += (
                        "Try setting the mode of the prediction client to "
                        "'auto' or 'online' to trigger inference upon request."
                    )
            else:
                msg += f" Server responded with error message: {str(error)}"
            result = None
        return result, msg

    def get_image_prediction(self, image: Image) -> Prediction:
        """
        Get a prediction for an image from the Intel® Geti™ server, if available.

        NOTE: This method is only available for images that are already existing on
        the server! For getting predictions on a 'new' image, please see the
        `PredictionClient.predict_image` method

        :param image: Image to get the prediction for. The image has to be present in
            the project on the cluster already.
        :return: Prediction for the image
        """
        result, msg = self._get_prediction_for_media_item(
            media_item=image, prediction_mode=self.mode
        )
        if result is None:
            raise ValueError(msg)
        return result

    def get_video_frame_prediction(self, video_frame: VideoFrame) -> Prediction:
        """
        Get a prediction for a video frame from the Intel® Geti™ server, if available.

        :param video_frame: VideoFrame to get the prediction for. The frame has to be
            present in the project on the cluster already.
        :return: Prediction for the video frame
        """
        result, msg = self._get_prediction_for_media_item(
            media_item=video_frame, prediction_mode=self.mode
        )
        if result is None:
            raise ValueError(msg)
        return result

    def get_video_predictions(self, video: Video) -> List[Prediction]:
        """
        Get a list of predictions for a video from the Intel® Geti™ server, if available.

        :param video: Video to get the predictions for. The video has to be present in
            the project on the cluster already.
        :return: List of Predictions for the video
        """
        result, msg = self._get_prediction_for_media_item(
            media_item=video, prediction_mode=self.mode
        )
        if result is None:
            raise ValueError(msg)
        return result

    def download_predictions_for_images(
        self,
        images: MediaList[Image],
        path_to_folder: str,
        include_result_media: bool = True,
    ) -> float:
        """
        Download image predictions from the server to a target folder on disk.

        :param images: List of images for which to download the predictions
        :param path_to_folder: Folder to save the predictions to
        :param include_result_media: True to also download the result media belonging
            to the predictions, if any. False to skip downloading result media
        :return: Returns the time elapsed to download the predictions, in seconds
        """
        return self._download_predictions_for_2d_media_list(
            media_list=images,
            path_to_folder=path_to_folder,
            include_result_media=include_result_media,
        )

    def download_predictions_for_videos(
        self,
        videos: MediaList[Video],
        path_to_folder: str,
        include_result_media: bool = True,
        inferred_frames_only: bool = True,
        frame_stride: Optional[int] = None,
    ) -> float:
        """
        Download predictions for a list of videos from the server to a target folder
        on disk.

        :param videos: List of videos for which to download the predictions
        :param path_to_folder: Folder to save the predictions to
        :param include_result_media: True to also download the result media belonging
            to the predictions, if any. False to skip downloading result media
        :param inferred_frames_only: True to only download frames that already have
            a prediction, False to run inference on the full video for all videos in
            the list.
            WARNING: Setting this to False may cause the download to take a long time!
        :param frame_stride: Optional frame stride to use when generating predictions.
            This is only used when `inferred_frames_only = False`. If left unspecified,
            the frame_stride is deduced from the video
        :return: Time elapsed to download the predictions, in seconds
        """
        t_total = 0
        logging.info(
            f"Starting prediction download... saving predictions for "
            f"{len(videos)} videos to folder {path_to_folder}/predictions"
        )
        for video in videos:
            t_total += self.download_predictions_for_video(
                video=video,
                path_to_folder=path_to_folder,
                include_result_media=include_result_media,
                inferred_frames_only=inferred_frames_only,
                frame_stride=frame_stride,
            )
        logging.info(f"Video prediction download finished in {t_total:.1f} seconds.")
        return t_total

    def download_predictions_for_video(
        self,
        video: Video,
        path_to_folder: str,
        include_result_media: bool = True,
        inferred_frames_only: bool = True,
        frame_stride: Optional[int] = None,
    ) -> float:
        """
        Download video predictions from the server to a target folder on disk.

        :param video: Video for which to download the predictions
        :param path_to_folder: Folder to save the predictions to
        :param include_result_media: True to also download the result media belonging
            to the predictions, if any. False to skip downloading result media
        :param inferred_frames_only: True to only download frames that already have
            a prediction, False to run inference on the full video.
            WARNING: Setting this to False may cause the download to take a long time!
        :param frame_stride: Optional frame stride to use when generating predictions.
            This is only used when `inferred_frames_only = False`. If left unspecified,
            the frame_stride is deduced from the video
        :return: Returns the time elapsed to download the predictions, in seconds
        """
        if inferred_frames_only:
            predictions = self.get_video_predictions(video=video)
            frame_list = MediaList[VideoFrame](
                [
                    VideoFrame.from_video(
                        video=video, frame_index=prediction.media_identifier.frame_index
                    )
                    for prediction in predictions
                ]
            )
        else:
            stride = (
                frame_stride
                if frame_stride is not None and frame_stride > 0
                else video.media_information.frame_stride
            )
            frame_indices = range(0, video.media_information.frame_count, stride)
            frame_list = MediaList[VideoFrame](
                [
                    VideoFrame.from_video(video=video, frame_index=frame_index)
                    for frame_index in frame_indices
                ]
            )
            # Set the prediction mode to online to force inference on frames that don't
            # have a prediction yet
            self._override_mode(PredictionMode.ONLINE)
        if len(frame_list) > 0:
            result = self._download_predictions_for_2d_media_list(
                media_list=frame_list,
                path_to_folder=path_to_folder,
                verbose=False,
                include_result_media=include_result_media,
            )
        else:
            result = 0
        self._reset_override_mode()
        return result

    def _override_mode(self, mode: PredictionMode):
        """
        Temporarily override the prediction mode.

        :param mode: new prediction mode to use
        :return:
        """
        self.__override_mode = mode

    def _reset_override_mode(self):
        """
        Remove override of the prediction mode.

        :return:
        """
        self.__override_mode = None

    def _download_predictions_for_2d_media_list(
        self,
        media_list: Union[MediaList[Image], MediaList[VideoFrame]],
        path_to_folder: str,
        include_result_media: bool = True,
        verbose: bool = True,
    ) -> float:
        """
        Download predictions from the server to a target folder on disk.

        :param media_list: List of images or video frames to download the predictions
            for
        :param path_to_folder: Folder to save the predictions to
        :param include_result_media: True to also download the result media belonging
            to the predictions, if any. False to skip downloading result media
        :param verbose: True to print verbose output, False to run in silent mode
        :return: Returns the time elapsed to download the predictions, in seconds
        """
        if media_list.media_type == Image:
            media_name = "image"
            media_name_plural = "images"
        elif media_list.media_type == VideoFrame:
            media_name = "video frame"
            media_name_plural = "video frames"
        else:
            raise ValueError(
                "Invalid media type found in media_list, unable to download "
                "predictions."
            )

        if not path_to_folder.endswith("predictions"):
            path_to_predictions_folder = os.path.join(path_to_folder, "predictions")
        else:
            path_to_predictions_folder = path_to_folder

        if verbose:
            logging.info(
                f"Starting prediction download... saving predictions for "
                f"{len(media_list)} {media_name_plural} to folder "
                f"{path_to_predictions_folder}"
            )
        os.makedirs(path_to_predictions_folder, exist_ok=True, mode=0o770)
        t_start = time.time()
        download_count = 0
        skip_count = 0
        tqdm_prefix = "Downloading predictions"
        with logging_redirect_tqdm(tqdm_class=tqdm):
            for media_item in tqdm(media_list, desc=tqdm_prefix):
                prediction, msg = self._get_prediction_for_media_item(
                    media_item,
                    prediction_mode=self.mode,
                    include_explanation=include_result_media,
                )
                if prediction is None:
                    if verbose:
                        logging.info(
                            f"Unable to retrieve prediction for {media_name} "
                            f"{media_item.name}, with reason: {msg}. Skipping this "
                            f"{media_name}"
                        )
                    skip_count += 1
                    continue
                kind = prediction.kind
                if kind != AnnotationKind.PREDICTION:
                    if verbose:
                        logging.warning(
                            f"Received invalid prediction of kind {kind} for {media_name} "
                            f"with name{media_item.name}"
                        )
                    skip_count += 1
                    continue

                # Download result media belonging to the prediction, if required
                if prediction.has_result_media and include_result_media:
                    try:
                        result_media = prediction.get_result_media_data(self.session)
                    except GetiRequestException:
                        if verbose:
                            logging.info(
                                f"Unable to retrieve prediction result map for "
                                f"{media_name} '{media_item.name}'. Skipping"
                            )
                        result_media = None
                    if result_media is not None:
                        path_to_result_media_folder = os.path.join(
                            path_to_predictions_folder, "saliency_maps"
                        )
                        os.makedirs(
                            path_to_result_media_folder, exist_ok=True, mode=0o770
                        )
                        for result_medium in result_media:
                            result_media_path = os.path.join(
                                path_to_result_media_folder,
                                media_item.name
                                + "_"
                                + result_medium.friendly_name
                                + ".jpg",
                            )

                            os.makedirs(
                                os.path.dirname(result_media_path),
                                exist_ok=True,
                                mode=0o770,
                            )
                            with open(result_media_path, "wb") as f:
                                f.write(result_medium.data)

                # Convert prediction to json and save to file
                export_data = PredictionRESTConverter.to_dict(prediction)
                prediction_path = os.path.join(
                    path_to_predictions_folder, media_item.name + ".json"
                )

                os.makedirs(os.path.dirname(prediction_path), exist_ok=True, mode=0o770)
                with open(prediction_path, "w") as f:
                    json.dump(export_data, f, indent=4)
                download_count += 1
        t_elapsed = time.time() - t_start
        if download_count > 0:
            msg = (
                f"Downloaded {download_count} predictions to folder "
                f"{path_to_predictions_folder} in {t_elapsed:.1f} seconds."
            )
        else:
            msg = "No predictions were downloaded."
        if skip_count > 0:
            msg = (
                msg + f" Was unable to retrieve predictions for {skip_count} "
                f"{media_name_plural}, these {media_name_plural} were skipped."
            )
        if verbose:
            logging.info(msg)
        return t_elapsed

    def predict_image(
        self, image: Union[Image, np.ndarray, os.PathLike, str]
    ) -> Prediction:
        """
        Push an image to the Intel® Geti™ project and receive a prediction for it.

        Note that this method will not save the image to the project.

        :param image: Image object, filepath to an image or numpy array containing an
            image to get the prediction for
        :return: Prediction for the image
        """
        # Get image pixel data from input
        image_data: Optional[np.ndarray]
        image_name: Optional[str]
        if isinstance(image, Image):
            image_data = image.numpy
            if image_data is None:
                raise ValueError(
                    "An 'Image' object was passed for prediction, but the image does "
                    "not contain any pixel data. Please make sure that the pixel data "
                    "is loaded for the image by calling 'image.get_data()' with the "
                    "appropriate Geti session."
                )
            image_name = image.name
        elif isinstance(image, np.ndarray):
            image_data = image
            image_name = "numpy_image.jpg"
        elif isinstance(image, (os.PathLike, str)):
            image_data = None
            image_name = None
        else:
            raise TypeError(
                f"Received object 'image' of type {type(image)}, which is invalid. "
                f"Please either pass an 'Image' object, a numpy array or a filepath."
            )

        if image_data is None:
            image_io = open(image, "rb").read()
        else:
            image_io = io.BytesIO(cv2.imencode(".jpg", image_data)[1].tobytes())
            image_io.name = image_name

        url = f"{self._base_url}pipelines/active:predict"
        contenttype = "multipart"
        data = {"file": image_io}

        # make POST request
        response = self.session.get_rest_response(
            url=url,
            method="POST",
            contenttype=contenttype,
            data=data,
        )
        prediction = PredictionRESTConverter.from_dict(response)
        prediction.resolve_label_names_and_colors(labels=self._labels)
        return prediction



# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import os
import warnings
from typing import List, Optional

from geti_sdk.data_models import (
    Dataset,
    Image,
    Model,
    Project,
    Subset,
    TrainingDatasetStatistics,
    VideoFrame,
)
from geti_sdk.data_models.containers import MediaList
from geti_sdk.http_session import GetiSession
from geti_sdk.http_session.exception import GetiRequestException
from geti_sdk.rest_converters import MediaRESTConverter
from geti_sdk.utils import deserialize_dictionary


class DatasetClient:
    """
    Class to manage datasets for a certain Intel® Geti™ project.
    """

    def __init__(self, workspace_id: str, project: Project, session: GetiSession):
        self.session = session
        self.project = project
        self.workspace_id = workspace_id
        self.base_url = f"workspaces/{workspace_id}/projects/{project.id}/datasets"

    def create_dataset(self, name: str) -> Dataset:
        """
        Create a new dataset named `name` inside the project.

        :param name: Name of the dataset to create
        :return: The created dataset
        """
        request_data = {"name": name}
        response = self.session.get_rest_response(
            url=self.base_url,
            method="POST",
            data=request_data,
        )
        dataset = deserialize_dictionary(response, output_type=Dataset)
        self.project.datasets.append(dataset)
        return dataset

    def delete_dataset(self, dataset: Dataset) -> None:
        """
        Delete provided dataset inside the project.

        :param dataset: Dataset to delete
        """
        try:
            response = self.session.get_rest_response(
                url=self.base_url + f"/{dataset.id}",
                method="DELETE",
            )
        except GetiRequestException as error:
            if error.status_code == 404:
                warnings.warn(
                    f"Dataset with name `{dataset.name}` was not found in the project. "
                    "Please make sure that the dataset you are trying to delete exists."
                )
            else:
                raise error
        if isinstance(response, dict) and response.get("result", None) == "success":
            logging.info(f"Dataset `{dataset.name}` was successfully deleted.")
        else:
            logging.error(f"Failed to delete dataset `{dataset.name}`.")

    def get_all_datasets(self) -> List[Dataset]:
        """
        Query the Intel® Geti™ server to retrieve an up to date list of datasets in
        the project.

        :return: List of current datasets in the project
        """
        if self.project.id is None:
            raise ValueError(
                "Project ID is not defined, please make sure that the project you pass "
                "contains a valid ID"
            )
        response = self.session.get_rest_response(url=self.base_url, method="GET")
        datasets = [
            deserialize_dictionary(dataset_dict, Dataset)
            for dataset_dict in response["datasets"]
        ]
        self.project.datasets = datasets
        return datasets

    def get_dataset_statistics(self, dataset: Dataset) -> dict:
        """
        Retrieve the media and annotation statistics for a particular dataset
        """
        response = self.session.get_rest_response(
            url=f"{self.base_url}/{dataset.id}/statistics", method="GET"
        )
        return response

    def get_dataset_by_name(self, dataset_name: str) -> Dataset:
        """
        Retrieve a dataset by name

        :param dataset_name: Name of the dataset to retrieve
        :return: Dataset object
        """
        dataset: Optional[Dataset] = None
        for ds in self.get_all_datasets():
            if ds.name == dataset_name:
                dataset = ds
        if dataset is None:
            raise ValueError(
                f"Dataset named '{dataset_name}' was not found in project "
                f"'{self.project.name}'"
            )
        return dataset

    def has_dataset_subfolders(self, path_to_folder: str) -> bool:
        """
        Check if a project folder has it's media folders organized according to the
        datasets in the project

        :param path_to_folder: Path to the project folder to check
        :return: True if the media folders in the project folder tree are organized
            according to the datasets in the project.
        """
        media_folder_names = ["images", "videos"]
        result: bool = True
        for folder_name in media_folder_names:
            full_path = os.path.join(path_to_folder, folder_name)
            if not os.path.exists(full_path):
                continue
            result *= self._media_folder_has_dataset_subfolders(full_path)
        return result

    def _media_folder_has_dataset_subfolders(self, path_to_folder: str) -> bool:
        """
        Check if a folder with media has subfolders organized according to the
        projects' datasets

        :param path_to_folder: Path to the media folder to check for dataset subfolders
        :return: True if the folder contains subfolders named according to the
            datasets in the project, False otherwise
        """
        if not os.path.isdir(path_to_folder):
            return False
        content = os.listdir(path_to_folder)
        for dataset in self.project.datasets:
            if dataset.name not in content:
                return False
            if not os.path.isdir(os.path.join(path_to_folder, dataset.name)):
                return False
        return True

    def get_training_dataset_summary(self, model: Model) -> TrainingDatasetStatistics:
        """
        Return information concerning the training dataset for the `model`.
        This includes the number of images and video frames, and the statistics for
        the subset splitting (i.e. the number of training, test and validation
        images/video frames)

        :param model: Model to get the training dataset for
        :return: A `TrainingDatasetStatistics` object, containing the training dataset
            statistics for the model
        """
        ds_info = model.training_dataset_info
        dataset_storage_id = ds_info.get("dataset_storage_id", None)
        revision_id = ds_info.get("dataset_revision_id", None)
        if dataset_storage_id is None or revision_id is None:
            raise ValueError(
                f"Unable to fetch the required dataset information from the model. "
                f"Expected dataset and revision id's, got {ds_info} instead."
            )
        training_dataset = self.session.get_rest_response(
            url=f"{self.base_url}/{dataset_storage_id}/training_revisions/{revision_id}",
            method="GET",
        )
        return deserialize_dictionary(training_dataset, TrainingDatasetStatistics)

    def get_media_in_training_dataset(
        self, model: Model, subset: str = "training"
    ) -> Subset:
        """
        Return the media in the training dataset for the `model`, for
        the specified `subset`. Subset can be `training`, `validation` or `testing`.

        :param model: Model for which to get the media in the training dataset
        :param subset: The subset for which to return the media items. Can be either
            `training` (the default), `validation` or `testing`
        return: A `Subset` object, containing lists of `images` and `video_frames` in
            the requested `subset`
        :raises: ValueError if the DatasetClient is unable to fetch the required
            dataset information from the model
        """
        ds_info = model.training_dataset_info
        dataset_storage_id = ds_info.get("dataset_storage_id", None)
        revision_id = ds_info.get("dataset_revision_id", None)
        if dataset_storage_id is None or revision_id is None:
            raise ValueError(
                f"Unable to fetch the required dataset information from the model. "
                f"Expected dataset and revision id's, got {ds_info} instead."
            )
        post_data = {
            "condition": "and",
            "rules": [{"field": "subset", "operator": "equal", "value": subset}],
        }

        images: MediaList[Image] = MediaList([])
        video_frames: MediaList[VideoFrame] = MediaList([])

        next_page = f"{self.base_url}/{dataset_storage_id}/training_revisions/{revision_id}/media:query"
        while next_page:
            response = self.session.get_rest_response(
                url=next_page, method="POST", data=post_data
            )
            next_page = response.get("next_page", None)
            for item in response["media"]:
                if item["type"] == "image":
                    item.pop("annotation_scene_id", None)
                    item.pop("editor_name", None)
                    item.pop("roi_id", None)
                    image = MediaRESTConverter.from_dict(item, Image)
                    images.append(image)

                if item["type"] == "video":
                    video_id = item["id"]
                    next_frame_page = (
                        f"{self.base_url}/{dataset_storage_id}/training_revisions/"
                        f"{revision_id}/media/videos/{video_id}:query"
                    )
                    while next_frame_page:
                        frames_response = self.session.get_rest_response(
                            url=next_frame_page, method="POST", data=post_data
                        )
                        for frame in frames_response["media"]:
                            frame["video_id"] = video_id
                            video_frame = MediaRESTConverter.from_dict(
                                frame, VideoFrame
                            )
                            video_frames.append(video_frame)
                        next_frame_page = frames_response.get("next_page", None)

        return Subset(images=images, frames=video_frames, purpose=subset)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
import logging
from typing import Dict, List, Optional, Sequence, Union

import cv2
import numpy as np
from datumaro import Image
from datumaro.components.annotation import Bbox, Polygon

from geti_sdk.annotation_readers.base_annotation_reader import AnnotationReader
from geti_sdk.data_models import Annotation as SCAnnotation
from geti_sdk.data_models import TaskType
from geti_sdk.data_models.enums.task_type import GLOBAL_TASK_TYPES
from geti_sdk.data_models.media import MediaInformation
from geti_sdk.rest_converters import AnnotationRESTConverter
from geti_sdk.utils import generate_segmentation_labels, get_dict_key_from_value

from .datumaro_dataset import DatumaroDataset


class DatumAnnotationReader(AnnotationReader):
    """
    Class to read annotations using datumaro
    """

    _SUPPORTED_TASK_TYPES = [
        TaskType.DETECTION,
        TaskType.SEGMENTATION,
        TaskType.CLASSIFICATION,
        TaskType.INSTANCE_SEGMENTATION,
        TaskType.ROTATED_DETECTION,
        TaskType.ANOMALY_CLASSIFICATION,
        TaskType.ANOMALY_DETECTION,
        TaskType.ANOMALY_SEGMENTATION,
        TaskType.ANOMALY,
    ]

    def __init__(
        self,
        base_data_folder: str,
        annotation_format: str,
        task_type: Union[TaskType, str] = TaskType.DETECTION,
    ):
        super().__init__(
            base_data_folder=base_data_folder,
            annotation_format=annotation_format,
            task_type=task_type,
        )
        self.dataset = DatumaroDataset(
            dataset_format=annotation_format, dataset_path=base_data_folder
        )
        self._override_label_map: Optional[Dict[int, str]] = None
        self._applied_filters: List[Dict[str, Union[List[str], str]]] = []

    def prepare_and_set_dataset(
        self,
        task_type: Union[TaskType, str],
        previous_task_type: Optional[TaskType] = None,
    ) -> None:
        """
        Prepare the dataset for a specific `task_type`. This could involve for
        example conversion of annotation shapes.

        :param task_type: TaskType for which to prepare the dataset.
        :param previous_task_type: TaskType preceding the task to prepare the dataset
            for
        """
        if not isinstance(task_type, TaskType):
            task_type = TaskType(task_type)
        if task_type != self.task_type:
            logging.info(f"Task type changed to {task_type} for dataset")
            if task_type not in self._SUPPORTED_TASK_TYPES:
                raise ValueError(f"Unsupported task type {task_type}")
            new_dataset = DatumaroDataset(
                dataset_format=self.annotation_format, dataset_path=self.base_folder
            )
            self.task_type = task_type
            self.dataset = new_dataset
            for filter_parameters in self.applied_filters:
                self.filter_dataset(**filter_parameters)

        dataset = self.dataset.prepare_dataset(
            task_type=task_type, previous_task_type=previous_task_type
        )
        self.dataset.set_dataset(dataset)
        logging.info(f"Dataset is prepared for {task_type} task.")

    def convert_labels_to_segmentation_names(self) -> None:
        """
        Convert the label names in a dataset to '*_shape`, where `*` is
        the original label name. It can be used to generate unique label names for the
        segmentation task in a detection_to_segmentation project
        """
        segmentation_label_map: Dict[int, str] = {}
        label_names = list(self.datum_label_map.values())
        segmentation_label_names = generate_segmentation_labels(label_names)
        for datum_index, label_name in self.datum_label_map.items():
            label_index = label_names.index(label_name)
            segmentation_label_map.update(
                {datum_index: segmentation_label_names[label_index]}
            )
        self.override_label_map(segmentation_label_map)

    def get_all_label_names(self) -> List[str]:
        """
        Retrieve the list of labels names from a datumaro dataset.
        """
        return list(set(self.datum_label_map.values()))

    @property
    def datum_label_map(self) -> Dict[int, str]:
        """
        :return: Dictionary mapping the datumaro label id to the label name
        """
        if self._override_label_map is None:
            return self.dataset.label_mapping
        else:
            return self._override_label_map

    def override_label_map(self, new_label_map: Dict[int, str]):
        """
        Override the label map defined in the datumaro dataset
        """
        self._override_label_map = new_label_map

    def reset_label_map(self):
        """
        Reset the label map back to the original one from the datumaro dataset.
        """
        self._override_label_map = None

    def get_all_image_names(self) -> List[str]:
        """
        Return a list of image names in the dataset
        """
        return self.dataset.image_names

    def get_data(
        self,
        filename: str,
        label_name_to_id_mapping: dict,
        media_information: MediaInformation,
        preserve_shape_for_global_labels: bool = False,
    ) -> List[SCAnnotation]:
        """
        Return the annotation data for the dataset item corresponding to `filename`.

        :param filename: name of the item to get the annotation data for.
        :param label_name_to_id_mapping: mapping of label name to label id.
        :param media_information: MediaInformation object containing information
            (e.g. width, height) about the media item to upload the annotation for
        :param preserve_shape_for_global_labels: False to convert shapes for global
            tasks to full rectangles (required for classification like tasks in
            Intel® Geti™ projects), True to preserve such shapes. This parameter
            should be:

             - False when uploading annotations to a single task project
             - True when uploading annotations for a classification like task,
                following a local task in a task chain project.

        :return: List of Annotation objects containing all annotations for the given
            dataset item.
        """
        ds_item = self.dataset.get_item_by_id(filename)
        image_size = ds_item.media_as(Image).size
        annotation_list: List[SCAnnotation] = []
        labels = []

        # Remove duplicate annotations, datumaro does not check for this
        datum_annotations = [
            i
            for n, i in enumerate(ds_item.annotations)
            if i not in ds_item.annotations[:n]
        ]

        for annotation in datum_annotations:
            try:
                label_name = self.datum_label_map[annotation.label]
            except KeyError:
                # Label is not in the Intel® Geti™ project labels, move on to next
                # annotation for this dataset item.
                continue

            label_id = label_name_to_id_mapping.get(label_name)
            label = {"id": label_id, "probability": 1.0}
            if (
                self.task_type not in GLOBAL_TASK_TYPES
                or preserve_shape_for_global_labels
            ):
                if isinstance(annotation, Bbox):
                    x1 = float(annotation.points[0])
                    y1 = float(annotation.points[1])
                    x2 = float(annotation.points[2])
                    y2 = float(annotation.points[3])
                    shape = {
                        "type": "RECTANGLE",
                        "x": x1,
                        "y": y1,
                        "width": x2 - x1,
                        "height": y2 - y1,
                    }
                elif isinstance(annotation, Polygon):
                    if self.task_type == TaskType.ROTATED_DETECTION:
                        contour = np.array(
                            [
                                np.array([x, y], dtype=np.float32)
                                for x, y in zip(*[iter(annotation.points)] * 2)
                            ],
                            dtype=np.float32,
                        )
                        min_rect = cv2.minAreaRect(contour)
                        shape = {
                            "type": "ROTATED_RECTANGLE",
                            "x": min_rect[0][0],
                            "y": min_rect[0][1],
                            "width": min_rect[1][0],
                            "height": min_rect[1][1],
                            "angle": min_rect[2],
                        }
                    else:
                        points = [
                            {"x": float(x), "y": float(y)}
                            for x, y in zip(*[iter(annotation.points)] * 2)
                        ]
                        shape = {"type": "POLYGON", "points": points}
                else:
                    logging.warning(
                        f"Unsupported annotation type found: "
                        f"{type(annotation)}. Skipping..."
                    )
                    continue
                sc_annotation = AnnotationRESTConverter.annotation_from_dict(
                    {"labels": [label], "shape": shape}
                )
                annotation_list.append(sc_annotation)
            else:
                labels.append(label)

        if not preserve_shape_for_global_labels and self.task_type in GLOBAL_TASK_TYPES:
            shape = {
                "type": "RECTANGLE",
                "x": 0.0,
                "y": 0.0,
                "width": float(image_size[1]),
                "height": float(image_size[0]),
            }
            sc_annotation = AnnotationRESTConverter.annotation_from_dict(
                {"labels": labels, "shape": shape}
            )
            annotation_list.append(sc_annotation)
        return annotation_list

    @property
    def applied_filters(self) -> List[Dict[str, Union[List[str], str]]]:
        """
        Return a list of filters and their parameters that have been previously
        applied to the dataset.
        """
        return copy.deepcopy(self._applied_filters)

    def filter_dataset(self, labels: Sequence[str], criterion="OR") -> None:
        """
        Retain only those items with annotations in the list of labels passed.

        :param: labels     List of labels to filter on
        :param: criterion  Filter criterion, currently "OR" or "AND" are implemented
        """
        self.dataset.filter_items_by_labels(labels=labels, criterion=criterion)
        self._applied_filters.append({"labels": labels, "criterion": criterion})

    def group_labels(self, labels_to_group: List[str], group_name: str) -> None:
        """
        Group multiple labels into one. Grouping converts the list of labels into one
        single label named `group_name`.

        This method does not return anything, but instead overrides the label map for
        the datamaro dataset to account for the grouping.

        :param labels_to_group: List of labels names that should be grouped together
        :param group_name: Name of the resulting label
        :return:
        """
        label_keys = [
            get_dict_key_from_value(self.datum_label_map, label)
            for label in labels_to_group
        ]
        new_label_map = copy.deepcopy(self.datum_label_map)
        for key in label_keys:
            new_label_map[key] = group_name
        self.override_label_map(new_label_map=new_label_map)

    def get_annotation_stats(self) -> Dict[str, Dict[str, int]]:
        """
        Return the object and image counts per label in the dataset.

        :return: Dictionary containing label names as keys, and as values:
            - n_images: Number of images containing this label
            - n_objects: Number of independent objects with this label
        """
        label_statistics: Dict[str, Dict[str, int]] = {}
        for item in self.dataset.dataset:
            label_names = set(
                [
                    self.datum_label_map[annotation.label]
                    for annotation in item.annotations
                ]
            )
            object_counts: Dict[str, int] = {}
            for label in label_names:
                object_counts[label] = 0
            for annotation in item.annotations:
                label_name = self.datum_label_map[annotation.label]
                object_counts[label_name] += 1

            for label in label_names:
                label_stats = label_statistics.get(label, {})
                if not label_stats:
                    label_statistics[label] = {
                        "n_images": 1,
                        "n_objects": object_counts[label],
                    }
                else:
                    label_statistics[label]["n_images"] += 1
                    label_statistics[label]["n_objects"] += object_counts[label]
        return label_statistics



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
from typing import List, Optional, Union

from .constants import BASE_TEST_PATH


def get_sdk_fixtures() -> List[str]:
    """
    Returns the list of fixtures available to the SDK

    Recursively traverses the directory tree under the `tests/fixtures` directory.

    :return: list of fixture paths for pytest to import
    """
    fixtures_path = os.path.join(BASE_TEST_PATH, "fixtures")
    fixtures = _get_fixtures_from_folder(
        path_to_folder=fixtures_path, parent_folders_dot_path="tests"
    )
    return fixtures


def _get_fixtures_from_folder(
    path_to_folder: Union[str, os.PathLike],
    parent_folders_dot_path: str,
    fixtures: Optional[List[str]] = None,
) -> List[str]:
    """
    Collect fixtures from the folder at `path_to_folder`

    :param path_to_folder: Path to the folder in which to look for fixtures
    :param parent_folders_dot_path: Path (relative from the base `tests` directory` to
        the folder, in dot-notation
    :param fixtures: existing list of fixtures, in dot notation paths
    :return: list of dot notation paths corresponding to files containing fixtures
    """
    if fixtures is None:
        fixtures: List[str] = []
    for filename in os.listdir(path_to_folder):
        full_filepath = os.path.join(path_to_folder, filename)
        new_dot_path = f"{parent_folders_dot_path}.{os.path.basename(path_to_folder)}"
        if os.path.isfile(full_filepath):
            if filename.endswith(".py") and not filename.startswith("__"):
                fixtures.append(f"{new_dot_path}.{filename[0:-3]}")
        elif os.path.isdir(full_filepath):
            fixtures.extend(
                _get_fixtures_from_folder(
                    path_to_folder=full_filepath, parent_folders_dot_path=new_dot_path
                )
            )
    return fixtures



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from enum import Enum


class SdkTestMode(Enum):
    """
    This Enum represents the different modes available for running the tests. The
    available modes are:
        - ONLINE:  The tests are run against an actual Geti server. Real http
                     requests are being made but no cassettes are recorded
                     Use this mode to verify that the SDK data models are still up to
                     date with the current Geti REST contracts
        - OFFLINE: The tests are run using the recorded requests and responses. All
                     http requests are intercepted an no actual requests will be made.
                     Use this mode in a CI environment, or for fast testing of the SDK
                     logic
        - RECORD:  The tests are run against an actual Geti server. HTTP requests
                     are being made and all requests and responses are recorded to a new
                     set of cassettes for the SDK test suite. The old cassettes for the
                     tests are deleted.
    """

    ONLINE = "online"
    OFFLINE = "offline"
    RECORD = "record"



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `rest_converters` package contains methods for serializing and deserializing
entities, using the data models defined in the :py:mod:`~geti_sdk.data_models`
module.

The rest converters are used in the :py:mod:`~geti_sdk.rest_clients`.

Module contents
---------------

.. autoclass:: ProjectRESTConverter
   :members:
   :undoc-members:

.. autoclass:: MediaRESTConverter
   :members:

.. autoclass:: AnnotationRESTConverter
   :members:

.. autoclass:: PredictionRESTConverter
   :members:

.. autoclass:: ConfigurationRESTConverter
   :members:

.. autoclass:: ModelRESTConverter
   :members:

.. autoclass:: StatusRESTConverter
   :members:

.. autoclass:: JobRESTConverter
   :members:

.. autoclass:: TestResultRESTConverter
   :members:
"""
from .annotation_rest_converter import AnnotationRESTConverter
from .configuration_rest_converter import ConfigurationRESTConverter
from .job_rest_converter import JobRESTConverter
from .media_rest_converter import MediaRESTConverter
from .model_rest_converter import ModelRESTConverter
from .prediction_rest_converter import PredictionRESTConverter
from .project_rest_converter import ProjectRESTConverter
from .status_rest_converter import StatusRESTConverter
from .test_result_rest_converter import TestResultRESTConverter

__all__ = [
    "ProjectRESTConverter",
    "MediaRESTConverter",
    "AnnotationRESTConverter",
    "PredictionRESTConverter",
    "ConfigurationRESTConverter",
    "ModelRESTConverter",
    "StatusRESTConverter",
    "JobRESTConverter",
    "TestResultRESTConverter",
]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
from typing import List

import setuptools

with open("README.md", "r", encoding="utf-8") as readme_file:
    LONG_DESCRIPTION = readme_file.read()


def get_requirements(filename: str) -> List[str]:
    """
    Get the required packages from the `filename` specified.
    """
    package_dir = os.path.dirname(os.path.abspath(__file__))
    filepath = os.path.join(package_dir, "requirements", filename)
    required_packages: List[str] = []
    with open(filepath, "r", encoding="utf-8") as requirements_file:
        for line in requirements_file:
            requirement = line.strip()
            if requirement and not requirement.startswith(("#", "-f")):
                required_packages.append(requirement)
    return required_packages


with open("geti_sdk/_version.py", "r", encoding="utf-8") as init_file:
    for line in init_file:
        line = line.strip()
        if line.startswith("__version__"):
            VERSION = line.split("=")[1].strip().strip('"')

setuptools.setup(
    name="geti-sdk",
    version=VERSION,
    author="Intel OpenVINO",
    author_email="ludo.cornelissen@intel.com",
    description="Software Development Kit for the Intel® Geti™ platform",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="Copyright (C) 2022 Intel Corporation - All Rights Reserved. Licensed "
    "under the Apache License, Version 2.0 (the 'License'). See LICENSE file for "
    "more details.",
    url="https://github.com/openvinotoolkit/geti-sdk",
    project_urls={
        "Documentation": "https://openvinotoolkit.github.io/geti-sdk",
        "Bug Tracker": "https://github.com/openvinotoolkit/geti-sdk/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.9,<3.12",
    install_requires=get_requirements("requirements.txt"),
    extras_require={
        "dev": get_requirements("requirements-dev.txt"),
        "docs": get_requirements("requirements-docs.txt"),
        "notebooks": get_requirements("requirements-notebooks.txt"),
    },
    include_package_data=True,
)



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import datetime
import inspect
import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List, Optional

import numpy as np

from geti_sdk.data_models import Prediction


class PostInferenceObject(object, metaclass=ABCMeta):
    """
    Base interface class for post inference triggers, actions and hooks
    """

    _override_from_dict_: bool = False

    def __new__(cls, *args, **kwargs):
        """
        Create a new PostInferenceObject instance
        """
        instance = super().__new__(cls)
        instance._argument_dict_ = {"args": args, "kwargs": kwargs}
        return instance

    def __init__(self):
        self._constructor_arguments_ = self.__get_constructor_arguments()

    def __get_constructor_arguments(self) -> Dict[str, Any]:
        """
        Return the arguments used for constructing the PostInferenceAction object

        :return: Dictionary containing the constructor parameter names as keys, and
            parameter values as values
        """
        constructor_argument_params = inspect.signature(self.__init__).parameters
        parameters: Dict[str, Any] = {}
        args = self._argument_dict_.get("args", ())
        kwargs = self._argument_dict_.get("kwargs", {})
        for index, (pname, parameter) in enumerate(constructor_argument_params.items()):
            if index + 1 <= len(args):
                parameters.update({pname: args[index]})
            elif pname in kwargs.keys():
                parameters.update({pname: kwargs[pname]})
            else:
                parameters.update({pname: parameter.default})
        return parameters

    def to_dict(self) -> Dict[str, Any]:
        """
        Return a dictionary representation of the PostInferenceObject

        :return: Dictionary representing the class name and its constructor parameters
        """
        constructor_args: Dict[str, Any] = {}
        for key, value in self._constructor_arguments_.items():
            if isinstance(value, PostInferenceObject):
                constructor_args.update({key: value.to_dict()})
            else:
                constructor_args.update({key: value})
        return {type(self).__name__: constructor_args}

    @classmethod
    def from_dict(cls, input_dict: Dict[str, Any]) -> "PostInferenceObject":
        """
        Construct a PostInferenceObject from an input dictionary `input_dict`

        :param input_dict: Dictionary representation of the PostInferenceObject
        :return: Instantiated PostInferenceObject, according to the input dictionary
        """
        available_objects = {subcls.__name__: subcls for subcls in cls.__subclasses__()}
        pi_objects: List["PostInferenceObject"] = []
        for object_name, object_args in input_dict.items():
            target_object = available_objects[object_name]
            if target_object._override_from_dict_:
                pi_objects.append(target_object.from_dict(object_args))
            else:
                pi_objects.append(target_object(**object_args))
        return pi_objects[0]


class PostInferenceTrigger(PostInferenceObject):
    """
    Base class for post inference triggers

    Post inference triggers are used in inference hooks. They define a condition
    (or set of conditions) that determines whether any post-inference action should be
    taken.

    :param threshold: Threshold to use in activating the trigger. If the computed
        score is above the threshold, the trigger emits a positive decision and any
        downstream actions will be performed
    """

    def __init__(self, threshold: float = 0.5):
        super().__init__()
        self.threshold = threshold
        self._repr_info_ = f"threshold={threshold:.1f}"

    @abstractmethod
    def __call__(self, image: np.ndarray, prediction: Prediction) -> float:
        """
        Compute a trigger score for the `image` and corresponding `prediction`.

        :param image: Numpy array representing an image
        :param prediction: Prediction object corresponding to the inference result
            for the image.
        :return: Float representing the score for the input
        """
        return NotImplemented

    def get_decision(self, score: float) -> bool:
        """
        Make a decision based on a previously computed `score` and the threshold defined
        for the trigger

        :param score: Float representing the score for a certain image, prediction pair
        :return: Boolean indicating whether the trigger conditions are met (True,
            score is higher than the threshold), or not (False)
        """
        return score > self.threshold

    def __repr__(self) -> str:
        """
        Return string representation of the PostInferenceTrigger
        """
        return f"{type(self).__name__}({self._repr_info_})"


class PostInferenceAction(PostInferenceObject):
    """
    Base class for post inference actions. These are actions that are used in inference
    hooks, and can be (conditionally) executed after inference

    :param log_level: Log level to use in the action, current options are `info` or
        `debug`. Defaults to `debug`
    """

    def __init__(self, log_level: str = "debug"):
        super().__init__()
        if log_level.lower() == "debug":
            self.log_function = logging.debug
        elif log_level.lower() == "info":
            self.log_function = logging.info
        else:
            raise ValueError(
                f"Unsupported log_level `{log_level}`, options are 'info' or 'debug'."
            )
        self._repr_info_: str = ""

    @abstractmethod
    def __call__(
        self,
        image: np.ndarray,
        prediction: Prediction,
        score: Optional[float] = None,
        name: Optional[str] = None,
        timestamp: Optional[datetime.datetime] = None,
    ):
        """
        Execute the action for the given `image` with corresponding `prediction` and an
        `score`. The `score` could be passed from the trigger that triggers the action.

        :param image: Numpy array representing an image
        :param prediction: Prediction object which was generated for the image
        :param score: Optional score computed from a post inference trigger
        :param name: String containing the name of the image
        :param timestamp: Datetime object containing the timestamp belonging to the
            image
        """
        return NotImplemented

    def __repr__(self) -> str:
        """
        Return string representation of the PostInferenceAction
        """
        return f"{type(self).__name__}({self._repr_info_})"


class PostInferenceHookInterface(PostInferenceObject):
    """
    Basic hook that can be executed after inference. A PostInferenceHook consists of:
      - A Trigger, defining a condition
      - An Action, defining an act or sequence of actions to perform when the
        trigger is activated.


    :param trigger: PostInferenceTrigger which is evaluated and used to trigger the
        hook action
    :param action: PostInferenceAction to be executed if triggered
    """

    def __init__(self, trigger: PostInferenceTrigger, action: PostInferenceAction):
        super().__init__()
        self.trigger = trigger
        self.action = action

    @abstractmethod
    def run(
        self,
        image: np.ndarray,
        prediction: Prediction,
        name: Optional[str] = None,
        timestamp: Optional[datetime.datetime] = None,
    ) -> None:
        """
        Execute the post inference hook. This will first evaluate the `trigger`, and
        if the trigger conditions are met the corresponding `action` will be executed.

        :param image: Numpy array representing the image that was the input for
            model inference
        :param prediction: Prediction object containing the inference results for the
            image
        :param name: Optional name of the image which can be used in the hook
            action, for example as filename or tag for data collection
        :param timestamp: Optional timestamp belonging to the image
        """
        raise NotImplementedError



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import os
import warnings
from glob import glob
from typing import Dict, List, Optional, Sequence, Set, Union

from geti_sdk.annotation_readers import AnnotationReader
from geti_sdk.data_models import Annotation, ScoredLabel, TaskType
from geti_sdk.data_models.enums.media_type import SUPPORTED_IMAGE_FORMATS
from geti_sdk.data_models.media import MediaInformation
from geti_sdk.data_models.shapes import Rectangle


class DirectoryTreeAnnotationReader(AnnotationReader):
    """
    AnnotationReader for loading single label classification annotations from a
    dataset organized in a directory tree. This annotation reader expects images to
    be put in folders, where the name of each image folder corresponds to the label
    that should be assigned to all images inside it.

    :param base_data_folder: Root of the directory tree that contains the dataset
    :param subset_folder_names: Optional list of subfolders of the base_data_folder
        that should not be considered as labels, but should be used to acquire the
        data. For example ['train', 'validation', 'test'] for a dataset that is split
        into three subsets.
    :param task_type: TaskType for the task in the Intel® Geti™ platform to which the
        annotations should be uploaded
    """

    def __init__(
        self,
        base_data_folder: str,
        subset_folder_names: Optional[Sequence[str]] = None,
        task_type: Union[TaskType, str] = TaskType.CLASSIFICATION,
    ):
        if not task_type.is_global:
            raise ValueError(
                "The DirectoryTreeAnnotationReader only supports annotations for "
                "global task types."
            )
        super().__init__(
            base_data_folder=base_data_folder,
            annotation_format="directory",
            task_type=task_type,
        )

        has_root_level_subsets = True if subset_folder_names is not None else False
        if has_root_level_subsets:
            self.target_data_dirs = [
                os.path.join(base_data_folder, subset) for subset in subset_folder_names
            ]
        else:
            self.target_data_dirs = [base_data_folder]
        # Label map is a dictionary mapping the root labels to new label names. It
        # can be used to filter or group the labels in the dataset
        self.has_filters_or_grouping = False
        self._original_labels = self.get_all_label_names()
        self._label_map: Dict[str, str] = {
            label: label for label in self._original_labels
        }

    @property
    def label_map(self) -> Dict[str, str]:
        """
        Return the label map for the dataset, mapping the root label names (keys) to
        potential new label names (values). It is used to filter or group the dataset.

        If no filters or grouping has been applied, it returns a dictionary with key,
        value pairs that have identical keys and values, i.e. {"dog": "dog"}
        """
        return self._label_map

    def reset_filters_and_grouping(self):
        """
        Reset the applied filters and grouping, to recover the original dataset
        """
        self._label_map = {label: label for label in self._original_labels}
        self.has_filters_or_grouping = False

    def get_data(
        self,
        filename: str,
        label_name_to_id_mapping: dict,
        media_information: MediaInformation,
        preserve_shape_for_global_labels: bool = False,
        image_name_as_full_path: bool = False,
    ) -> List[Annotation]:
        """
        Return the list of annotations for the media item with name `filename`

        :param filename: Name of the item to return the annotations for
        :param label_name_to_id_mapping: Dictionary mapping the name of a label to its
            unique database ID
        :param media_information: MediaInformation object containing information
            (e.g. width, height) about the media item to upload the annotation for
        :param preserve_shape_for_global_labels: Unused parameter in this type of
            annotation reader
        :param image_name_as_full_path: Set to True if the `filename` contains the
            full path to the image
        :return: A list of Annotation objects for the media item
        """
        filepath = ""
        annotations: List[Annotation] = []
        if image_name_as_full_path:
            label_matches = [os.path.basename(os.path.dirname(filename))]
            extension = filename[:-4]
            if extension in SUPPORTED_IMAGE_FORMATS:
                filepath = filename
            else:
                for format_extension in SUPPORTED_IMAGE_FORMATS:
                    full_name = filename + format_extension
                    if os.path.isfile(full_name):
                        filepath = full_name
                if filepath == "":
                    raise ValueError(
                        f"No valid image file found at path {filename}, unable to "
                        f"generate annotation data."
                    )
        else:
            matches = glob(
                os.path.join(self.base_folder, "**", f"{filename}.*"), recursive=True
            )
            label_matches = [
                os.path.basename(os.path.dirname(match)) for match in matches
            ]
            if len(label_matches) > 1:
                warnings.warn(
                    f"Multiple matching labels found for image with "
                    f"name {filename}: {label_matches}. Skipping this image..."
                )
            elif len(label_matches) == 0:
                logging.info(
                    f"Image with name {filename} was not found in the dataset at path "
                    f"{self.base_folder}. Skipping this image..."
                )
                return []
            filepath = matches[0]
        label_name = self.label_map[label_matches[0]]
        label = ScoredLabel(
            name=label_name,
            probability=1.0,
            id=label_name_to_id_mapping[label_name],
        )
        annotations.append(
            Annotation(
                labels=[label],
                shape=Rectangle(
                    x=0,
                    y=0,
                    width=media_information.width,
                    height=media_information.height,
                ),
            )
        )
        return annotations

    def get_all_label_names(self) -> List[str]:
        """
        Identify all label names contained in the dataset
        """
        label_names: Set[str] = set()
        for directory in self.target_data_dirs:
            for path, sub_directories, files in os.walk(directory):
                for sub_directory in sub_directories:
                    if self.has_filters_or_grouping:
                        if sub_directory not in self.label_map.keys():
                            continue
                    if self.has_filters_or_grouping:
                        label_name = self.label_map[sub_directory]
                    else:
                        label_name = sub_directory
                    label_names.add(label_name)
        return list(label_names)

    def get_data_filenames(self) -> List[str]:
        """
        Return a list of annotated media files found in the dataset.

        :return: List of filenames (excluding extension) for all annotated files in
            the data folder
        """
        data_file_paths: List[str] = []
        for directory in self.target_data_dirs:
            for path, sub_directories, files in os.walk(directory):
                for name in files:
                    if self.has_filters_or_grouping:
                        if os.path.basename(path) not in self.label_map.keys():
                            continue
                    data_file_paths.append(
                        os.path.join(path, os.path.splitext(name)[0])
                    )
        return data_file_paths

    def filter_dataset(self, labels: Sequence[str], criterion: str = "OR") -> None:
        """
        Retain only those items with annotations in the list of labels passed.

        :param labels: List of labels to filter on
        :param criterion: Unused parameter for this type of annotation reader
        """
        self._label_map = {label: label for label in labels}
        self.applied_filters.append({"labels": labels, "criterion": criterion})
        self.has_filters_or_grouping = True

    def group_labels(self, labels_to_group: List[str], group_name: str) -> None:
        """
        Group multiple labels into one. Grouping converts the list of labels into one
        single label named `group_name`.

        This method does not return anything, but instead overrides the label map for
        the annotation reader to account for the grouping.

        :param labels_to_group: List of labels names that should be grouped together
        :param group_name: Name of the resulting label
        """
        for org_label_name, mapped_label in self.label_map.items():
            if org_label_name in labels_to_group:
                self._label_map.update({org_label_name: group_name})
        self.has_filters_or_grouping = True

    def get_annotation_stats(self) -> Dict[str, Dict[str, int]]:
        """
        Return the image counts per label in the dataset.

        :return: Dictionary containing label names as keys, and as values:
            - n_images: Number of images containing this label
        """
        label_statistics: Dict[str, Dict[str, int]] = {}
        label_names = self.get_all_label_names()
        for label in label_names:
            label_statistics[label] = {"n_images": 0}
        for item_filepath in self.get_data_filenames():
            item_label = self.label_map[
                os.path.basename(os.path.dirname(item_filepath))
            ]
            label_statistics[item_label]["n_images"] += 1
        return label_statistics



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import datetime
import glob
import io
import logging
import os
from typing import List, Optional, Sequence, Union

import cv2
import numpy as np

from geti_sdk.data_models import Dataset, Image, MediaType
from geti_sdk.data_models.containers import MediaList
from geti_sdk.rest_converters import MediaRESTConverter

from .media_client import MEDIA_SUPPORTED_FORMAT_MAPPING, BaseMediaClient


class ImageClient(BaseMediaClient[Image]):
    """
    Class to manage image uploads and downloads for a certain project.
    """

    _MEDIA_TYPE = MediaType.IMAGE

    def get_all_images(self, dataset: Optional[Dataset] = None) -> MediaList[Image]:
        """
        Get the ID's and filenames of all images in the project, from a specific
        dataset. If no dataset is passed, images from the training dataset will be
        returned

        :param dataset: Dataset for which to retrieve the images. If no dataset is
            passed, images from the training dataset are returned.
        :return: MediaList containing all Image entities in the dataset
        """
        return self._get_all(dataset=dataset)

    def upload_image(
        self,
        image: Union[np.ndarray, str, os.PathLike],
        dataset: Optional[Dataset] = None,
    ) -> Image:
        """
        Upload an image file to the server.

        :param image: full path to the image on disk, or numpy array representing the
            image
        :param dataset: Dataset to which to upload the image. If no dataset is
            passed, the image is uploaded to the training dataset
        :return: Image object representing the uploaded image on the server
        """
        if isinstance(image, (str, os.PathLike)):
            image_dict = self._upload(image, dataset=dataset)
        elif isinstance(image, np.ndarray):
            image_io = io.BytesIO(cv2.imencode(".jpg", image)[1].tobytes())
            time_now = datetime.datetime.now()
            image_io.name = f"numpy_{time_now.strftime('%Y-%m-%dT%H-%M-%S.%f')}.jpg"
            image_dict = self._upload_bytes(image_io, dataset=dataset)
        else:
            raise TypeError(f"Invalid image type: {type(image)}.")
        return MediaRESTConverter.from_dict(input_dict=image_dict, media_type=Image)

    def upload_folder(
        self,
        path_to_folder: str,
        n_images: int = -1,
        skip_if_filename_exists: bool = False,
        dataset: Optional[Dataset] = None,
        max_threads: int = 5,
    ) -> MediaList[Image]:
        """
        Upload all images in a folder to the project. Returns a MediaList containing
        all images in the project after upload.

        :param path_to_folder: Folder with images to upload
        :param n_images: Number of images to upload from folder
        :param skip_if_filename_exists: Set to True to skip uploading of an image
            if an image with the same filename already exists in the project.
            Defaults to False
        :param dataset: Dataset to which to upload the images. If no dataset is
            passed, the images are uploaded to the training dataset
        :param max_threads: Maximum number of threads to use for uploading. Defaults to 5.
            Set to -1 to use all available threads.
        :return: MediaList containing all image's in the project
        """
        return self._upload_folder(
            path_to_folder=path_to_folder,
            n_media=n_images,
            skip_if_filename_exists=skip_if_filename_exists,
            dataset=dataset,
            max_threads=max_threads,
        )

    def download_all(
        self,
        path_to_folder: str,
        append_image_uid: bool = False,
        max_threads: int = 10,
        dataset: Optional[Dataset] = None,
    ) -> None:
        """
        Download all images in a project or a dataset to a folder on the local disk.

        :param path_to_folder: path to the folder in which the images should be saved
        :param append_image_uid: True to append the UID of an image to the
            filename (separated from the original filename by an underscore, i.e.
            '{filename}_{image_id}'). If there are images in the project/dataset with
            duplicate filename, this must be set to True to ensure all images are
            downloaded. Otherwise, images with the same name will be skipped.
        :param max_threads: Maximum number of threads to use for downloading. Defaults to 10.
            Set to -1 to use all available threads.
        :param dataset: Dataset from which to download the images. If no dataset is provided,
            images from all datasets are downloaded.
        """
        self._download_all(
            path_to_folder,
            append_media_uid=append_image_uid,
            max_threads=max_threads,
            dataset=dataset,
        )

    def upload_from_list(
        self,
        path_to_folder: str,
        image_names: List[str],
        extension_included: bool = False,
        n_images: int = -1,
        skip_if_filename_exists: bool = False,
        image_names_as_full_paths: bool = False,
        dataset: Optional[Dataset] = None,
        max_threads: int = 5,
    ) -> MediaList[Image]:
        """
        From a folder containing images `path_to_folder`, this method uploads only
        those images that have their filenames included in the `image_names` list.

        :param path_to_folder: Folder containing the images
        :param image_names: List of names of the images that should be uploaded
        :param extension_included: Set to True if the extension of the image is
            included in the name, for each image in the image_names list. Defaults to
            False
        :param n_images: Number of images to upload from the list
        :param skip_if_filename_exists: Set to True to skip uploading of an image
            if an image with the same filename already exists in the project.
            Defaults to False
        :param image_names_as_full_paths: Set to True if the list of `image_names`
            contains full paths to the images, rather than just the filenames
        :param dataset: Dataset to which to upload the images. If no dataset is
            passed, the images are uploaded to the training dataset
        :param max_threads: Maximum number of threads to use for uploading images.
            Defaults to 5. Set to -1 to use all available threads.
        :return: List of images that were uploaded
        """
        media_formats = MEDIA_SUPPORTED_FORMAT_MAPPING[self._MEDIA_TYPE]

        if n_images > len(image_names) or n_images == -1:
            n_to_upload = len(image_names)
        else:
            n_to_upload = n_images

        image_filepaths: List[str] = []
        if image_names_as_full_paths:
            if extension_included:
                image_filepaths = image_names
            else:
                for image_name in image_names:
                    for media_extension in media_formats:
                        if os.path.isfile(image_name + media_extension):
                            image_filepaths.append(image_name + media_extension)
                            break
            image_filepaths = image_filepaths[0:n_to_upload]

        else:
            logging.debug("Retrieving full filepaths for image upload...")
            filenames_lookup = {
                os.path.basename(path): path
                for path in glob.glob(
                    os.path.join(path_to_folder, "**"), recursive=True
                )
            }
            for image_name in image_names[0:n_to_upload]:
                matches: List[str] = []
                if not extension_included:
                    for media_extension in media_formats:
                        if (
                            filename := f"{image_name}{media_extension}"
                        ) in filenames_lookup:
                            matches.append(filenames_lookup[filename])
                else:
                    matches = (
                        [filenames_lookup[image_name]]
                        if image_name in filenames_lookup
                        else []
                    )
                if not matches:
                    raise ValueError(
                        f"No matching file found for image with name {image_name}"
                    )
                elif len(matches) != 1:
                    raise ValueError(
                        f"Multiple files found for image with name {image_name}: "
                        f"{matches}"
                    )
                image_filepaths.append(matches[0])
        return self._upload_loop(
            filepaths=image_filepaths,
            skip_if_filename_exists=skip_if_filename_exists,
            dataset=dataset,
            max_threads=max_threads,
        )

    def delete_images(self, images: Sequence[Image]) -> bool:
        """
        Delete all Image entities in `images` from the project.

        :param images: List of Image entities to delete
        :return: True if all images on the list were deleted successfully,
            False otherwise
        """
        return self._delete_media(media_list=images)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import time
import warnings
from datetime import datetime
from functools import cache
from json import JSONDecodeError
from typing import Any, Dict, Optional, Union

import requests
import simplejson
import urllib3
from requests import Response
from requests.exceptions import RequestException
from requests.structures import CaseInsensitiveDict
from urllib3.exceptions import InsecureRequestWarning

from geti_sdk.platform_versions import GetiVersion

from .exception import GetiRequestException
from .server_config import ServerCredentialConfig, ServerTokenConfig

CSRF_COOKIE_NAME = "_oauth2_proxy_csrf"
PROXY_COOKIE_NAME = "_oauth2_proxy"
GETI_COOKIE_NAME = "geti-cookie"

# INITIAL_HEADERS = {"Connection": "keep-alive", "Upgrade-Insecure-Requests": "1"}
INITIAL_HEADERS = {"Upgrade-Insecure-Requests": "1"}
SUCCESS_STATUS_CODES = [200, 201, 202, 204]

SAAS_MODE = "saas"
ONPREM_MODE = "on-prem"


class GetiSession(requests.Session):
    """
    Wrapper for requests.session that sets the correct headers and cookies, and
    handles authentication and authorization.

    :param server_config: Server configuration holding the hostname (or ip address) of
        the Intel® Geti™ server, as well as the details required for authentication
        (either username and password or personal access token)
    """

    def __init__(
        self,
        server_config: Union[ServerTokenConfig, ServerCredentialConfig],
    ):
        super().__init__()
        self.headers.update(INITIAL_HEADERS)
        self.allow_redirects = False
        self.token = None
        self._cookies: Dict[str, Optional[str]] = {
            CSRF_COOKIE_NAME: None,
            PROXY_COOKIE_NAME: None,
        }

        # Configure proxies
        if server_config.proxies is None:
            self._proxies: Dict[str, str] = {}
        else:
            self._proxies = server_config.proxies

        # Configure certificate verification
        if not server_config.has_valid_certificate:
            warnings.warn(
                "You have disabled TLS certificate validation, HTTPS requests made to "
                "the Intel® Geti™ server may be compromised. For optimal security, "
                "please enable certificate validation.",
                InsecureRequestWarning,
            )
            urllib3.disable_warnings(InsecureRequestWarning)
        self.verify = server_config.has_valid_certificate

        self.config = server_config
        self.logged_in = False
        self._auth_service: Optional[str] = None

        # Determine authentication method
        if isinstance(server_config, ServerCredentialConfig):
            if self.platform_serving_mode == SAAS_MODE:
                raise ValueError(
                    "Authentication via username and password is not supported for "
                    "Intel® Geti™ SaaS instances. Please use a personal access token."
                )
            else:
                logging.warning(
                    "Authentication via username and password is deprecated and will be "
                    "removed in a future version of the SDK. Please use a personal access token."
                )
            self.authenticate_with_password(verbose=True)
            self.use_token = False
        else:  # ServerTokenConfig
            self.use_token = True
            # New mechanism, from Geti v1.14
            self.headers.update({"x-api-key": f"{server_config.token}"})
            self.headers.pop("Connection")

        # Get server version
        self._product_info = self._get_product_info_and_set_api_version()
        self._organization_id: Optional[str] = self._get_organization_id()

    @property
    @cache
    def platform_serving_mode(self) -> str:
        """
        Return the type of the GETi platform service.
        """
        try:
            deployment_config_response = self.request(
                url=f"{self.config.host}/deployment-config.json",
                method="GET",
                proxies=self._proxies,
            ).json()
            serving_mode = deployment_config_response.get("servingMode").lower()
        except requests.exceptions.JSONDecodeError:
            return ONPREM_MODE
        if serving_mode == "on-prem":
            return ONPREM_MODE
        elif serving_mode == "saas":
            return SAAS_MODE
        else:
            raise ValueError(
                f"Unexpected serving mode '{serving_mode}' received from the server."
            )

    @property
    def version(self) -> GetiVersion:
        """
        Return the version of the Intel® Geti™ platform that is running on the server.

        :return: Version object holding the Intel® Geti™ version number
        """
        if "build-version" in self._product_info.keys():
            version_string = self._product_info.get("build-version")
        else:
            version_string = self._product_info.get("product-version")
        return GetiVersion(version_string)

    @property
    def organization_id(self) -> str:
        """
        Return the organization id for the user who is associated with the GetiSession
        """
        if self._organization_id is None:
            self._organization_id = self._get_organization_id()
        return self._organization_id

    def _get_initial_login_url(self) -> str:
        """
        Retrieve the initial login url by making a request to the login page, and
        following the redirects.

        :return: string containing the URL to the login page
        """
        params = (
            "client_id=web_ui"
            "&redirect_uri=/callback"
            "&code_challenge_method=S256"
            "&response_type=code"
            "&response_mode=query"
            "&scope=openid+profile+groups+email+offline_access"
        )
        url = f"{self.config.host}/dex/auth/regular_users?{params}"
        response = self.get(url, allow_redirects=False, proxies=self._proxies)
        if response.status_code in [302, 303]:
            login_page_url = response.next.url
        else:
            login_page_url = response.url
        return login_page_url

    def authenticate_with_password(self, verbose: bool = True):
        """
        Get a new authentication cookie from the server.

        :param verbose: True to print progress output, False to suppress output
        """
        if self.logged_in:
            logging.info("Already logged in, authentication is skipped")
            return
        self.cookies.clear()
        try:
            login_path = self._get_initial_login_url()
        except requests.exceptions.SSLError as error:
            raise requests.exceptions.SSLError(
                f"Connection to Intel® Geti™ server at '{self.config.host}' failed, "
                f"the server address can be resolved but the SSL certificate could not "
                f"be verified. \n Full error description: {error.args[-1]}"
            )
        except requests.exceptions.ConnectionError as error:
            if "dummy" in self.config.password or "dummy" in self.config.username:
                raise ValueError(
                    "Connection to the Intel® Geti™ server failed, please make sure to "
                    "update the user login information for the Intel® Geti™ instance."
                ) from error
            raise ValueError(
                f"Connection to the Intel® Geti™ server at host '{self.config.host}' "
                f"failed, please provide a valid cluster hostname or ip address as"
                f" well as valid login details."
            ) from error
        self.headers.clear()
        self.headers.update({"Content-Type": "application/x-www-form-urlencoded"})
        if verbose:
            logging.info(f"Authenticating on host {self.config.host}...")
        cookies = {}
        headers = self.headers
        response = self.post(
            url=login_path,
            data={"login": self.config.username, "password": self.config.password},
            cookies=cookies,
            headers=headers,
            allow_redirects=True,
            proxies=self._proxies,
        )
        self._handle_dex_response(response)
        if verbose:
            logging.info("Authentication successful. Cookie received.")
        self.logged_in = True

    def get_rest_response(
        self,
        url: str,
        method: str,
        contenttype: str = "json",
        data: Optional[Any] = None,
        allow_reauthentication: bool = True,
        include_organization_id: bool = True,
        allow_text_response: bool = False,
        request_headers: Dict[str, str] = {},
    ) -> Union[Response, dict, list]:
        """
        Return the REST response from a request to `url` with `method`.

        :param url: the REST url without the hostname and api pattern
        :param method: 'GET', 'POST', 'PUT', 'DELETE'
        :param contenttype: currently either 'json', 'jpeg', 'multipart', 'zip', or '',
            defaults to "json"
        :param data: the data to send in a post request, as json
        :param allow_reauthentication: True to handle authentication errors
            by attempting to re-authenticate. If set to False, such errors
            will be raised instead.
        :param include_organization_id: True to include the organization ID in the base
            URL. Can be set to False for accessing certain internal endpoints that do
            not require an organization ID, but do require error handling.
        :param allow_text_response: False to trigger error handling when the server
            returns a response with text/html content. This can happen in some cases
            when authentication has expired. However, some endpoints are designed to
            return text responses, for those endpoints this parameter should be set to
            True
        :param request_headers: Additional headers to include in the request
        """
        if self.config.api_pattern in url:
            url = url.split(self.config.api_pattern)[-1]

        self._update_headers_for_content_type(content_type=contenttype)

        if not include_organization_id or url.startswith(
            f"organizations/{self.organization_id}/"
        ):
            requesturl = f"{self.config.base_url}{url}"
        else:
            requesturl = f"{self.base_url}{url}"

        if method == "POST" or method == "PUT":
            if contenttype == "json":
                kw_data_arg = {"json": data}
            elif contenttype == "multipart":
                kw_data_arg = {"files": data}
            elif contenttype == "jpeg" or contenttype == "zip":
                kw_data_arg = {"data": data}
            else:
                raise ValueError(
                    f"Making a POST request with content of type {contenttype} is "
                    f"currently not supported through the Intel Geti SDK."
                )
        elif method == "PATCH":
            kw_data_arg = {"data": data}
        else:
            kw_data_arg = {}

        request_params = {
            "method": method,
            "url": requesturl,
            **kw_data_arg,
            "stream": True,
        }

        if not self.use_token:
            request_params.update({"cookies": self._cookies})

        csrf_header_methods = ["POST", "PUT", "DELETE", "PATCH"]
        if method in csrf_header_methods:
            self.headers.update({"x-geti-csrf-protection": "1"})
        else:
            self.headers.pop("x-geti-csrf-protection", "")

        # Make the request, retrying a maximum of 5 times in case of connection errors
        retries = 5
        last_conn_error: Optional[ConnectionError] = None
        while retries:
            try:
                response = self.request(
                    **request_params, proxies=self._proxies, headers=request_headers
                )
                break
            except requests.exceptions.SSLError as error:
                raise requests.exceptions.SSLError(
                    f"Connection to Intel® Geti™ server at '{self.config.host}' failed, "
                    f"the server address can be resolved but the SSL certificate could not "
                    f"be verified. \n Full error description: {error.args[-1]}"
                )
            except ConnectionError as conn_error:
                last_conn_error = conn_error
                retries -= 1

        if last_conn_error is not None:
            raise last_conn_error
        response_content_type = response.headers.get("Content-Type", [])
        if response.status_code not in SUCCESS_STATUS_CODES or (
            "text/html" in response_content_type and not allow_text_response
        ):
            response = self._handle_error_response(
                response=response,
                request_params=request_params,
                request_data=kw_data_arg,
                allow_reauthentication=allow_reauthentication,
                content_type=contenttype,
            )
        if response.headers.get("Content-Type", "").startswith("application/json"):
            result = response.json()
        else:
            result = response

        return result

    def logout(self, verbose: bool = True) -> None:
        """
        Log out of the server and end the session. All HTTPAdapters are closed and
        cookies and headers are cleared.

        :param verbose: False to suppress output messages
        """
        requires_closing = True
        requires_sign_out = True

        if not self.logged_in or self.use_token:
            requires_closing = False
            requires_sign_out = False

        if requires_sign_out:
            sign_out_url = (
                self.config.base_url[: -len(self.config.api_pattern)]
                + "/oauth2/sign_out"
            )
            try:
                response = self.request(
                    url=sign_out_url, method="GET", proxies=self._proxies
                )

                if response.status_code in SUCCESS_STATUS_CODES:
                    if verbose:
                        logging.info("Logout successful.")
                else:
                    raise GetiRequestException(
                        method="GET",
                        url=sign_out_url,
                        status_code=response.status_code,
                        request_data={},
                    )
            except (RequestException, AttributeError):
                if verbose:
                    logging.info(
                        f"The {self.__class__.__name__} is closed successfully, but "
                        f"the Geti instance was not able to logout from the server."
                    )
            except TypeError:
                # The session is already closed
                requires_closing = False

        self._cookies = {}
        self.cookies.clear()
        self.headers = CaseInsensitiveDict(INITIAL_HEADERS)
        self.logged_in = False
        if requires_closing:
            try:
                super().close()
            except TypeError as error:
                # Sometimes a TypeError is raised if logout is called during garbage
                # collection, however we can safely ignore this since the session will
                # be deleted in that case anyway.
                logging.debug(f"{error} encountered during GetiSession closure.")

    def _get_product_info_and_set_api_version(self) -> Dict[str, str]:
        """
        Return the product info as retrieved from the Intel® Geti™ server.

        This method will also attempt to set the API version correctly, based on the
        retrieved product info.

        :return: Dictionary containing the product info.
        """
        product_info = self.get_rest_response(
            "product_info", "GET", include_organization_id=False
        )
        return product_info

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Log out of the Intel® Geti™ server. This method is called when exiting the
        runtime context related to the session, in case the session is used as a context
        manager.
        """
        if self.logged_in:
            try:
                self.logout(verbose=False)
            except Exception as exc:
                logging.debug(f"{exc} encountered while exiting GetiSession context.")
        super().__exit__(exc_type, exc_value, traceback)

    def __del__(self):
        """
        Log out of the Intel® Geti™ server. This method is called when the session is
        deleted from memory.
        """
        if self.logged_in:
            try:
                self.logout(verbose=False)
            except Exception as exc:
                logging.debug(f"{exc} encountered during GetiSession deletion.")

    def _handle_error_response(
        self,
        response: Response,
        request_params: Dict[str, Any],
        request_data: Dict[str, Any],
        allow_reauthentication: bool = True,
        content_type: str = "json",
    ) -> Response:
        """
        Handle error responses from the server.

        :param response: The Response object received from the server
        :param request_params: Dictionary containing the original parameters of the
            request
        :param allow_reauthentication: True to handle authentication errors
            by attempting to re-authenticate. If set to False, such errors
            will be raised instead.
        :param content_type: The content type of the original request
        :raises: GetiRequestException in case the error cannot be handled
        :return: Response object resulting from the request
        """
        retry_request = False

        if response.status_code in [200, 401] and allow_reauthentication:
            # Authentication has likely expired, re-authenticate
            logging.info("Authentication may have expired, re-authenticating...")
            self.logged_in = False
            if not self.use_token:
                self.authenticate_with_password(verbose=False)
                logging.info("Authentication complete.")

            else:
                raise ValueError(
                    "Authentication via your personal access token has failed, "
                    "most likely the token has expired. Please verify that you "
                    "have provided a valid token."
                )

            retry_request = True

        elif response.status_code == 503:
            # In case of Service Unavailable, wait some time and try again. If it
            # still doesn't work, raise exception
            time.sleep(1)
            retry_request = True

        # We make one attempt to do the request again. If it fails again, a
        # GetiRequestException will be raised holding further details of the
        # reason for failure.
        if retry_request:
            self._update_headers_for_content_type(content_type=content_type)
            # Reset any file buffers that were included in the request data, so that we
            # can attempt to upload them again.
            if content_type == "multipart":
                for file_name, file_buffer in request_params["files"].items():
                    file_buffer.seek(0, 0)

            response = self.request(**request_params, proxies=self._proxies)

            if response.status_code in SUCCESS_STATUS_CODES:
                return response

        try:
            response_data = response.json()
        except (JSONDecodeError, simplejson.errors.JSONDecodeError):
            response_data = None

        raise GetiRequestException(
            method=request_params["method"],
            url=request_params["url"],
            status_code=response.status_code,
            request_data=request_data,
            response_data=response_data,
        )

    def _update_headers_for_content_type(self, content_type: str) -> None:
        """
        Update the session headers to contain the correct content type

        :param content_type: content type for the request
        """
        if content_type == "json":
            self.headers.update({"Content-Type": "application/json"})
        elif content_type == "jpeg":
            self.headers.update({"Content-Type": "image/jpeg"})
        elif content_type == "multipart":
            self.headers.pop("Content-Type", None)
        elif content_type == "":
            self.headers.pop("Content-Type", None)
        elif content_type == "zip":
            self.headers.update({"Content-Type": "application/zip"})
        elif content_type == "offset+octet-stream":
            self.headers.update({"Content-Type": "application/offset+octet-stream"})
        else:
            self.headers.update({"Content-Type": content_type})

    @property
    def base_url(self) -> str:
        """
        Return the base URL to the Intel Geti server. If the server is running
        Geti v1.9 or later, the organization ID will be included in the URL
        """
        return f"{self.config.base_url}organizations/{self.organization_id}/"

    def _get_organization_id(self) -> str:
        """
        Return the organization ID associated with the user and host information configured
        in this Session.

        NOTE: When authenticating with username and password, this method returns the
        ID of the default organization for the user!
        """
        if not self.use_token:
            result = self.get_rest_response(
                url="profile",
                method="GET",
                include_organization_id=False,
            )
        else:
            result = self.get_rest_response(
                url="personal_access_tokens/organization",
                method="GET",
                include_organization_id=False,
            )
        if "organizationId" in result.keys():
            # Geti < 2.5, return the id directly
            org_id = result.get("organizationId", None)
        elif "organizations" in result.keys():
            # Geti 2.5 and up: list of organizations, return the default one (the one
            # which was created at the earliest time)
            org_list = result["organizations"]
            creation_times = [
                datetime.fromisoformat(
                    x["organizationCreatedAt"].replace("Z", "+00:00")
                )
                for x in org_list
            ]
            ids = [x["organizationId"] for x in org_list]
            earliest_creation_time = min(creation_times)
            earliest_idx = creation_times.index(earliest_creation_time)
            org_id = ids[earliest_idx]
        else:
            org_id = None

        if org_id is None:
            raise ValueError(
                f"Unable to retrieve organization ID from the Intel Geti server. "
                f"Server responded with: `{result}`"
            )
        return org_id

    def _handle_legacy_dex_response(self, response: Response) -> None:
        """
        Process the login response for the old-style DEX authorization mechanism

        :param response: Response object to handle
        """
        try:
            previous_response = response.history[-1]
        except IndexError:
            raise ValueError(
                "The cluster responded to the request, but authentication failed. "
                "Please verify that you have provided correct credentials."
            )
        proxy_cookie = previous_response.cookies.get(PROXY_COOKIE_NAME)
        if proxy_cookie is not None:
            cookie = {PROXY_COOKIE_NAME: proxy_cookie}
            self._cookies.update(cookie)
        else:
            logging.warning(
                f"Authentication appears to have failed! No valid oauth cookie "
                f"obtained. Invalid response received from server. Status code: "
                f"{response.status_code}"
            )

    def _handle_dex_response(self, response: Response) -> None:
        """
        Process the login response from the DEX authorization mechanism

        Extract the code and exchange it for a token, which must then
        be used as authentication cookie

        :param response: DEX response object to handle
        """
        if response.status_code == 200:
            code = response.url.split("code=")[1].split("&")[0]
        elif response.status_code == 401:
            raise ValueError(
                "The cluster responded to the request, but authentication failed. "
                "Please verify that you have provided correct credentials."
            )
        elif response.status_code == 404:
            raise ValueError(
                "Unable to authenticate with the Intel Geti server. The authentication "
                "mechanism you are trying to use is no longer supported. This error "
                "indicates that the Intel® Geti™ server version is not supported by "
                "this version of the Intel Geti SDK package. Please update the "
                "Intel® Geti™ server to version 2.0 or later, or use a previous "
                "version of the SDK."
            )
        else:
            raise GetiRequestException(
                method=response.request.method,
                url=response.url,
                status_code=response.status_code,
                request_data={},
                response_data={
                    "message": response.content,
                    "error_code": response.reason,
                },
            )
        data = {
            "grant_type": "authorization_code",
            "redirect_uri": "/callback",
            "code": code,
            "client_id": "web_ui",
        }
        login_response = self.post(
            url=f"{self.config.host}/dex/token",
            data=data,
            allow_redirects=True,
            proxies=self._proxies,
        )
        token = login_response.json().get("access_token", None)
        self._cookies.update({GETI_COOKIE_NAME: token})



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import json
import logging
import os
from datetime import datetime
from typing import Optional

import cv2
import numpy as np

from geti_sdk.data_models import Prediction
from geti_sdk.deployment.inference_hook_interfaces import PostInferenceAction
from geti_sdk.prediction_visualization.visualizer import Visualizer
from geti_sdk.rest_converters import PredictionRESTConverter


class FileSystemDataCollection(PostInferenceAction):
    """
    Post inference action that will save an image to a specified folder on disk. The
    prediction output and trigger score that triggered the action are also saved.

    The data is saved in the `target_folder`, in which the action will create the
    following folder structure:

    <target_folder>
        |
        |- images
        |- predictions
        |- scores

    :param target_folder: Target folder on disk where the inferred images should be
        saved. If it does not exist yet, this action will create it.
    :param file_name_prefix: Prefix to use for the files that will be saved by this
        action. Default is 'image'
    :param save_predictions: True to save the predictions for each image as well, in
        addition to the image itself. Set to False to not store any predictions
    :param save_scores: True to save the trigger score for each image as well, in
        addition to the image itself. Set to False to not save any scores.
    :param log_level: Log level for the action. Options are 'info' or 'debug'
    """

    def __init__(
        self,
        target_folder: str,
        file_name_prefix: str = "",
        save_predictions: bool = True,
        save_scores: bool = True,
        save_overlays: bool = True,
        log_level: str = "debug",
    ):
        super().__init__(log_level=log_level)

        self.image_path = os.path.join(target_folder, "images")
        folders_to_create = [self.image_path]
        if save_predictions:
            self.predictions_path = os.path.join(target_folder, "predictions")
            folders_to_create.append(self.predictions_path)
        if save_scores:
            self.scores_path = os.path.join(target_folder, "scores")
            folders_to_create.append(self.scores_path)
        if save_overlays:
            self.overlays_path = os.path.join(target_folder, "overlays")
            folders_to_create.append(self.overlays_path)

        for path in folders_to_create:
            os.makedirs(path, exist_ok=True, mode=0o770)

        self.prefix = file_name_prefix
        self.save_predictions = save_predictions
        self.save_scores = save_scores
        self.save_overlays = save_overlays
        self.visualizer = Visualizer()

        self._repr_info_ = (
            f"target_folder=`{target_folder}`, "
            f"file_prefix={file_name_prefix}, "
            f"save_predictions={save_predictions}, "
            f"save_scores={save_scores}, "
            f"save_overlays={save_overlays}"
        )

    def __call__(
        self,
        image: np.ndarray,
        prediction: Prediction,
        score: Optional[float] = None,
        name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
    ):
        """
        Execute the action, save the given `image` to the predefined target folder.
        The `prediction` and `score` are also saved.

        :param image: Numpy array representing an image
        :param prediction: Prediction object which was generated for the image
        :param score: Optional score computed from a post inference trigger
        :param name: String containing the name of the image
        :param timestamp: Datetime object containing the timestamp belonging to the
            image
        """
        image_bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        # we use cv2 to encode the numpy array as image, so it expects an
        # image in BGR format. However, `Deployment.infer` requires RGB format, so
        # we have to convert
        if name is None:
            name = self.prefix
        else:
            name = self.prefix + name
        if timestamp is None:
            timestamp = datetime.now()

        filename = name + "_" + timestamp.strftime("%Y%m%dT%H-%M-%S-%f")

        success = cv2.imwrite(
            os.path.join(self.image_path, filename + ".png"), image_bgr
        )
        if not success:
            logging.error(
                f"Failed to save image `{filename}.png` to folder `{self.image_path}`."
            )

        try:
            if self.save_predictions:
                prediction_filepath = os.path.join(
                    self.predictions_path, filename + ".json"
                )
                with open(prediction_filepath, "w") as file:
                    prediction_dict = PredictionRESTConverter.to_dict(prediction)
                    json.dump(prediction_dict, fp=file)

            if self.save_scores:
                if score is not None:
                    score_filepath = os.path.join(self.scores_path, filename + ".txt")
                    with open(score_filepath, "w") as file:
                        file.write(f"{score}")

            if self.save_overlays:
                overlay_path = os.path.join(self.overlays_path, filename + ".jpg")
                result = self.visualizer.draw(image, prediction)
                self.visualizer.save_image(result, overlay_path)
        except Exception as e:
            logging.exception(e, stack_info=True, exc_info=True)

        self.log_function(
            f"FileSystemDataCollection inference action saved image data to folder "
            f"`{self.image_path}`"
        )
